from __future__ import annotations

import math
from collections.abc import Callable, MutableSequence, Sequence
from typing import TYPE_CHECKING

import msgspec

from grim.geom import Vec2

from ...creatures.damage_types import CreatureDamageType
from ...creatures.lifecycle import creature_lifecycle_is_alive, creature_lifecycle_is_collidable
from ...effects import EffectPool
from ...math_parity import NATIVE_HALF_PI, f32
from ...owner_ref import OwnerRef
from ...perks import PerkId
from ...weapons import weapon_entry_for_projectile_type_id
from ..types import (
    MAIN_PROJECTILE_POOL_SIZE,
    CreatureDamageApplier,
    Projectile,
    ProjectileCollisionProfile,
    ProjectileHit,
    ProjectileRuntimeState,
    ProjectileTemplateId,
)
from .behaviors import (
    _PROJECTILE_HIT_PERK_HOOKS,
    _ProjectileHitInfo,
    _ProjectileHitPerkCtx,
    _ProjectileUpdateCtx,
)
from .collision import _apply_damage_to_creature, _hit_radius_for, _within_native_find_radius
from .primary_rules import apply_linger, apply_post_hit, apply_pre_hit, primary_rule_for_type_id
from .spatial_hash import CreatureSpatialHash

if TYPE_CHECKING:
    from ...creatures.runtime import CreatureState
    from ...sim.state_types import PlayerState


class ProjectileUpdateOptions(msgspec.Struct, frozen=True):
    world_size: float
    damage_scale_by_type: dict[int, float]
    rng: Callable[[], int]
    runtime_state: ProjectileRuntimeState
    players: Sequence[PlayerState]
    apply_player_damage: Callable[[int, float], None]
    ion_aoe_scale: float = 1.0
    detail_preset: int = 5
    on_hit: Callable[[ProjectileHit], object] | None = None
    on_hit_post: Callable[[ProjectileHit, object], None] | None = None


class PrimaryStepCtx(msgspec.Struct, frozen=True):
    dt: float
    creatures: Sequence[CreatureState]
    options: ProjectileUpdateOptions


_DEFAULT_PROJECTILE_COLLISION_PROFILE = ProjectileCollisionProfile(
    hit_radius=1.0,
    initial_damage_pool=1.0,
)

_PROJECTILE_COLLISION_PROFILE_BY_TYPE_ID: dict[ProjectileTemplateId, ProjectileCollisionProfile] = {
    ProjectileTemplateId.ION_MINIGUN: ProjectileCollisionProfile(hit_radius=3.0, initial_damage_pool=1.0),
    ProjectileTemplateId.ION_RIFLE: ProjectileCollisionProfile(hit_radius=5.0, initial_damage_pool=1.0),
    ProjectileTemplateId.ION_CANNON: ProjectileCollisionProfile(hit_radius=10.0, initial_damage_pool=1.0),
    ProjectileTemplateId.PLASMA_CANNON: ProjectileCollisionProfile(hit_radius=10.0, initial_damage_pool=1.0),
    ProjectileTemplateId.GAUSS_GUN: ProjectileCollisionProfile(hit_radius=1.0, initial_damage_pool=300.0),
    ProjectileTemplateId.FIRE_BULLETS: ProjectileCollisionProfile(hit_radius=1.0, initial_damage_pool=240.0),
    ProjectileTemplateId.BLADE_GUN: ProjectileCollisionProfile(hit_radius=1.0, initial_damage_pool=50.0),
}


def projectile_collision_profile(type_id: ProjectileTemplateId) -> ProjectileCollisionProfile:
    return _PROJECTILE_COLLISION_PROFILE_BY_TYPE_ID.get(
        type_id,
        _DEFAULT_PROJECTILE_COLLISION_PROFILE,
    )


class ProjectilePool:
    def __init__(self, *, size: int = MAIN_PROJECTILE_POOL_SIZE) -> None:
        self._entries = [Projectile() for _ in range(size)]
        self._creature_damage_applier: CreatureDamageApplier | None = None

    @property
    def entries(self) -> list[Projectile]:
        return self._entries

    @property
    def creature_damage_applier(self) -> CreatureDamageApplier | None:
        return self._creature_damage_applier

    @creature_damage_applier.setter
    def creature_damage_applier(self, value: CreatureDamageApplier | None) -> None:
        self._creature_damage_applier = value

    def reset(self) -> None:
        for entry in self._entries:
            entry.active = False

    def spawn(
        self,
        *,
        pos: Vec2,
        angle: float,
        type_id: ProjectileTemplateId,
        owner: OwnerRef,
        travel_budget: float = 0.0,
        hits_players: bool = False,
    ) -> int:
        index = None
        for i, entry in enumerate(self._entries):
            if not entry.active:
                index = i
                break
        if index is None:
            index = len(self._entries) - 1
        entry = self._entries[index]

        entry.active = True
        # Native projectile spawn writes angle/pos as float32 fields; keep those
        # stores narrowed so next-tick movement uses the same precision.
        angle_f32 = float(f32(float(angle)))
        pos_f32 = Vec2(float(f32(float(pos.x))), float(f32(float(pos.y))))
        entry.angle = angle_f32
        entry.pos = pos_f32
        entry.origin = pos_f32
        entry.vel = Vec2(
            float(f32(math.cos(float(angle_f32)) * 1.5)),
            float(f32(math.sin(float(angle_f32)) * 1.5)),
        )
        entry.type_id = type_id
        entry.life_timer = 0.4
        entry.reserved = 0.0
        entry.speed_scale = 1.0
        entry.travel_budget = float(travel_budget)
        weapon_entry = weapon_entry_for_projectile_type_id(type_id)
        entry.travel_budget = float(weapon_entry.travel_budget)
        entry.owner = owner
        entry.hits_players = bool(hits_players)

        collision_profile = projectile_collision_profile(type_id)
        entry.hit_radius = float(collision_profile.hit_radius)
        entry.damage_pool = float(collision_profile.initial_damage_pool)
        return index

    def iter_active(self) -> list[Projectile]:
        return [entry for entry in self._entries if entry.active]

    def step(self, ctx: PrimaryStepCtx) -> list[ProjectileHit]:
        """Update the main projectile pool.

        Modeled after `projectile_update` (0x00420b90) for the subset used by demo/state-9 work.
        """
        dt = float(f32(float(ctx.dt)))
        creatures = ctx.creatures
        options = ctx.options
        world_size = float(f32(float(options.world_size)))
        damage_scale_by_type = options.damage_scale_by_type
        ion_aoe_scale = float(options.ion_aoe_scale)
        detail_preset = int(options.detail_preset)
        rng = options.rng
        runtime_state = options.runtime_state
        players = options.players
        apply_player_damage = options.apply_player_damage
        apply_creature_damage = self._creature_damage_applier
        on_hit = options.on_hit
        on_hit_post = options.on_hit_post

        if dt <= 0.0:
            return []

        barrel_greaser_active = False
        ion_gun_master_active = False
        ion_scale = float(ion_aoe_scale)
        poison_idx = int(PerkId.POISON_BULLETS)
        barrel_idx = int(PerkId.BARREL_GREASER)
        ion_idx = int(PerkId.ION_GUN_MASTER)
        for player in players:
            perk_counts = player.perk_counts

            if 0 <= barrel_idx < len(perk_counts) and int(perk_counts[barrel_idx]) > 0:
                barrel_greaser_active = True
            if 0 <= ion_idx < len(perk_counts) and int(perk_counts[ion_idx]) > 0:
                ion_gun_master_active = True
            if barrel_greaser_active and ion_gun_master_active:
                break

        if ion_scale == 1.0 and ion_gun_master_active:
            ion_scale = 1.2

        def _owner_perk_active(owner: OwnerRef, perk_idx: int) -> bool:
            player_index = owner.player_index_in_bounds(len(players))
            if player_index is None:
                return False
            perk_counts = players[player_index].perk_counts
            return 0 <= perk_idx < len(perk_counts) and int(perk_counts[perk_idx]) > 0

        effects: EffectPool | None = runtime_state.effects
        sfx_queue: MutableSequence[str] | None = runtime_state.sfx_queue

        hits: list[ProjectileHit] = []
        margin = 64.0

        def _creature_is_collidable(creature: CreatureState) -> bool:
            if not creature.active:
                return False
            if not creature_lifecycle_is_collidable(creature.lifecycle_stage):
                return False
            return True

        creature_spatial = CreatureSpatialHash(creatures=creatures, is_collidable=_creature_is_collidable)

        def _damage_scale(type_id: int) -> float:
            value = damage_scale_by_type.get(type_id)
            if value is not None:
                return float(value)
            return float(weapon_entry_for_projectile_type_id(ProjectileTemplateId(type_id)).damage_scale)

        def _damage_distance_f32(origin: Vec2, pos: Vec2) -> float:
            dx = float(f32(float(origin.x) - float(pos.x)))
            dy = float(f32(float(origin.y) - float(pos.y)))
            dist_sq = float(f32(float(f32(float(dx) * float(dx))) + float(f32(float(dy) * float(dy)))))
            return float(f32(math.sqrt(float(dist_sq))))

        def _projectile_damage_amount_f32(dist: float, damage_scale: float) -> float:
            # `projectile_update` computes this from float locals (`fVar11/fVar23`),
            # so keep the damage path rounded through float32.
            dist_f32 = float(f32(float(dist)))
            if dist_f32 < 50.0:
                dist_f32 = 50.0
            damage_scale_f32 = float(f32(float(damage_scale)))
            return float(f32(((100.0 / float(dist_f32)) * float(damage_scale_f32) * 30.0 + 10.0) * 0.95))

        def _damage_type_for() -> int:
            return int(CreatureDamageType.BULLET)

        update_ctx = _ProjectileUpdateCtx(
            pool=self,
            creatures=creatures,
            dt=float(dt),
            ion_scale=float(ion_scale),
            detail_preset=int(detail_preset),
            rng=rng,
            runtime_state=runtime_state,
            effects=effects,
            sfx_queue=sfx_queue,
        )

        def _reset_shock_chain_if_owner(index: int) -> None:
            if runtime_state.shock_chain_projectile_id != index:
                return
            runtime_state.shock_chain_projectile_id = -1
            runtime_state.shock_chain_links_left = 0

        for proj_index, proj in enumerate(self._entries):
            if not proj.active:
                continue
            rule = primary_rule_for_type_id(ProjectileTemplateId(proj.type_id))

            if proj.life_timer <= 0.0:
                proj.active = False
                # Native `projectile_update` clears the active flag but still
                # runs this tick's life_timer branch, so expired ion projectiles
                # can apply one final linger AoE pass.

            if proj.life_timer < 0.4:
                if rule.reset_shock_chain_on_linger:
                    _reset_shock_chain_if_owner(proj_index)
                apply_linger(rule.linger, update_ctx, proj)
                continue

            if (
                proj.pos.x < -margin
                or proj.pos.y < -margin
                or proj.pos.x > world_size + margin
                or proj.pos.y > world_size + margin
            ):
                proj.life_timer = float(f32(float(proj.life_timer) - float(dt)))
                continue

            steps = int(proj.travel_budget)
            if barrel_greaser_active and proj.owner.is_player():
                steps *= 2

            # Decompile parity (`projectile_update`, 0x00420b90):
            #   local_cc += (float)(cos(angle - pi/2) * frame_dt * 20.0f) * speed_scale * 3.0f
            #   local_c8 += (float)(sin(angle - pi/2) * frame_dt * 20.0f) * speed_scale * 3.0f
            # Keep the float32 cast before `* speed_scale * 3.0`.
            heading_radians = float(proj.angle) - NATIVE_HALF_PI
            dir_x = math.cos(heading_radians)
            dir_y = math.sin(heading_radians)
            acc = Vec2()
            step = 0
            while step < steps:
                acc = Vec2(
                    float(
                        f32(
                            float(acc.x)
                            + float(
                                f32(float(dir_x) * float(dt) * 20.0)
                                * float(proj.speed_scale)
                                * 3.0,
                            ),
                        ),
                    ),
                    float(
                        f32(
                            float(acc.y)
                            + float(
                                f32(float(dir_y) * float(dt) * 20.0)
                                * float(proj.speed_scale)
                                * 3.0,
                            ),
                        ),
                    ),
                )

                if acc.length() >= 4.0 or steps <= step + 3:
                    move = acc
                    proj.pos = Vec2(
                        float(f32(float(proj.pos.x) + float(move.x))),
                        float(f32(float(proj.pos.y) + float(move.y))),
                    )
                    acc = Vec2()

                    hit_idx = None
                    owner_creature_idx = proj.owner.creature_index_in_bounds(len(creatures))
                    for idx in creature_spatial.candidate_indices(pos=proj.pos, radius=float(proj.hit_radius)):
                        creature = creatures[idx]
                        if not _creature_is_collidable(creature):
                            continue
                        if _within_native_find_radius(
                            origin=proj.pos,
                            target=creature.pos,
                            radius=float(proj.hit_radius),
                            target_size=float(creature.size),
                        ):
                            hit_idx = idx
                            break

                    owner_collision = hit_idx is not None and owner_creature_idx is not None and int(hit_idx) == owner_creature_idx
                    if owner_collision:
                        # Native `creature_find_in_radius` does not skip owner id during
                        # search; owner hits are discarded after the first match instead of
                        # continuing to a later candidate in the same tick.
                        hit_idx = None

                    if hit_idx is None:
                        can_hit_players = True
                        if int(proj_index) == int(
                            runtime_state.shock_chain_projectile_id,
                        ):
                            # Native skips `player_find_in_radius` for the currently tracked
                            # shock-chain projectile slot in this branch.
                            can_hit_players = False

                        if proj.hits_players and can_hit_players:
                            hit_player_idx = None
                            owner_player_index = proj.owner.player_index_in_bounds(len(players))
                            for idx, player in enumerate(players):
                                if owner_player_index is not None and idx == owner_player_index:
                                    continue
                                if float(player.health) <= 0.0:
                                    continue
                                if _within_native_find_radius(
                                    origin=proj.pos,
                                    target=player.pos,
                                    radius=float(proj.hit_radius),
                                    target_size=float(player.size),
                                ):
                                    hit_player_idx = idx
                                    break

                            if hit_player_idx is None:
                                step += 3
                                continue

                            proj.life_timer = 0.25
                            apply_player_damage(int(hit_player_idx), 10.0)

                            step += 3
                            continue

                        step += 3
                        continue

                    type_id = proj.type_id
                    creature = creatures[hit_idx]

                    perk_ctx = _ProjectileHitPerkCtx(
                        proj=proj,
                        creature=creature,
                        rng=rng,
                        owner_perk_active=_owner_perk_active,
                        poison_idx=poison_idx,
                    )
                    for hook in _PROJECTILE_HIT_PERK_HOOKS:
                        hook(perk_ctx)

                    apply_pre_hit(rule.pre_hit, update_ctx, proj, int(hit_idx))

                    owner_player_index = proj.owner.player_index_in_bounds(len(runtime_state.shots_hit))
                    if owner_player_index is not None and creature_lifecycle_is_alive(creature.lifecycle_stage):
                        shots_hit = runtime_state.shots_hit
                        shots_hit[owner_player_index] += 1

                    target = creature.pos
                    hit = ProjectileHit(
                        type_id=type_id,
                        origin=proj.origin,
                        hit=proj.pos,
                        target=target,
                    )
                    hits.append(hit)
                    hit_ctx = on_hit(hit) if on_hit is not None else None

                    if proj.life_timer != 0.25 and rule.stop_on_hit:
                        proj.life_timer = 0.25
                        jitter = rng() & 3
                        jitter_dx = float(f32(float(dir_x) * float(jitter)))
                        jitter_dy = float(f32(float(dir_y) * float(jitter)))
                        proj.pos = Vec2(
                            float(f32(float(proj.pos.x) + float(jitter_dx))),
                            float(f32(float(proj.pos.y) + float(jitter_dy))),
                        )

                    dist = _damage_distance_f32(proj.origin, proj.pos)

                    apply_post_hit(
                        rule.post_hit,
                        update_ctx,
                        _ProjectileHitInfo(
                            proj_index=int(proj_index),
                            proj=proj,
                            hit_idx=int(hit_idx),
                            move=move,
                            target=target,
                        ),
                    )

                    damage_scale = _damage_scale(type_id)
                    damage_amount = _projectile_damage_amount_f32(dist, damage_scale)

                    if damage_amount > 0.0 and creature.hp > 0.0:
                        remaining = proj.damage_pool - 1.0
                        proj.damage_pool = remaining
                        # Native `projectile_update` writes both impulse components from the
                        # same cosine term (`cos(angle - pi/2) * speed_scale`).
                        impulse_axis = f32(math.cos(float(proj.angle) - NATIVE_HALF_PI) * float(proj.speed_scale))
                        impulse = Vec2(float(impulse_axis), float(impulse_axis))
                        damage_type = _damage_type_for()
                        if remaining <= 0.0:
                            _apply_damage_to_creature(
                                creatures,
                                int(hit_idx),
                                float(damage_amount),
                                damage_type=damage_type,
                                impulse=impulse,
                                owner=proj.owner,
                                apply_creature_damage=apply_creature_damage,
                            )
                            creature_spatial.sync_index(int(hit_idx))
                            if proj.life_timer != 0.25:
                                proj.life_timer = 0.25
                        else:
                            _apply_damage_to_creature(
                                creatures,
                                int(hit_idx),
                                float(remaining),
                                damage_type=damage_type,
                                impulse=impulse,
                                owner=proj.owner,
                                apply_creature_damage=apply_creature_damage,
                            )
                            creature_spatial.sync_index(int(hit_idx))
                            proj.damage_pool -= float(creature.hp)

                    # Native `projectile_update` has projectile-type specific freeze-hit
                    # handling. Non-Gauss/non-Fire-Bullets impacts emit a single shard
                    # here; Gauss/Fire-Bullets emits shards inside the six-iteration
                    # large-streak loop (presentation hook parity).
                    if (
                        float(runtime_state.bonuses.freeze) > 0.0
                        and effects is not None
                        and rule.emit_freeze_shard
                    ):
                        shard_angle = float(float(proj.angle) - NATIVE_HALF_PI)
                        shard_angle += float(int(rng()) % 0x264) * 0.01
                        effects.spawn_freeze_shard(
                            pos=proj.pos,
                            angle=float(shard_angle),
                            rand=rng,
                            detail_preset=int(detail_preset),
                        )

                    if proj.damage_pool == 1.0:
                        # Native clears damage_pool to 0.0 whenever it's exactly 1.0
                        # in this branch, even if life_timer is already 0.25.
                        life_before = float(proj.life_timer)
                        proj.damage_pool = 0.0
                        if life_before != 0.25:
                            proj.life_timer = 0.25

                    if proj.life_timer == 0.25 and rule.stop_on_hit:
                        if on_hit_post is not None and hit_ctx is not None:
                            on_hit_post(hit, hit_ctx)
                        break

                    if proj.damage_pool <= 0.0:
                        if on_hit_post is not None and hit_ctx is not None:
                            on_hit_post(hit, hit_ctx)
                        break

                    if on_hit_post is not None and hit_ctx is not None:
                        on_hit_post(hit, hit_ctx)

                step += 3

        return hits

    def update_demo(
        self,
        dt: float,
        creatures: Sequence[CreatureState],
        *,
        world_size: float,
        speed_by_type: dict[int, float],
        damage_by_type: dict[int, float],
    ) -> list[ProjectileHit]:
        """Update a small projectile subset for the demo view."""

        if dt <= 0.0:
            return []

        hits: list[ProjectileHit] = []
        margin = 64.0

        for proj in self._entries:
            if not proj.active:
                continue

            if proj.life_timer <= 0.0:
                proj.active = False
                continue

            if proj.life_timer < 0.4:
                if proj.type_id == ProjectileTemplateId.ION_RIFLE:
                    damage = dt * 100.0
                    radius = 88.0
                    for creature in creatures:
                        if creature.hp <= 0.0:
                            continue
                        creature_radius = _hit_radius_for(creature)
                        hit_r = radius + creature_radius
                        if Vec2.distance_sq(proj.pos, creature.pos) <= hit_r * hit_r:
                            creature.hp -= damage
                elif proj.type_id == ProjectileTemplateId.ION_MINIGUN:
                    damage = dt * 40.0
                    radius = 60.0
                    for creature in creatures:
                        if creature.hp <= 0.0:
                            continue
                        creature_radius = _hit_radius_for(creature)
                        hit_r = radius + creature_radius
                        if Vec2.distance_sq(proj.pos, creature.pos) <= hit_r * hit_r:
                            creature.hp -= damage
                proj.life_timer = float(f32(float(proj.life_timer) - float(dt)))
                continue

            if (
                proj.pos.x < -margin
                or proj.pos.y < -margin
                or proj.pos.x > world_size + margin
                or proj.pos.y > world_size + margin
            ):
                proj.life_timer = float(f32(float(proj.life_timer) - float(dt)))
                continue

            speed = speed_by_type.get(proj.type_id, 650.0) * proj.speed_scale
            direction = Vec2.from_heading(float(proj.angle))
            proj.pos = proj.pos + direction * (speed * dt)

            hit_idx = None
            for idx, creature in enumerate(creatures):
                if creature.hp <= 0.0:
                    continue
                creature_radius = _hit_radius_for(creature)
                hit_r = proj.hit_radius + creature_radius
                if Vec2.distance_sq(proj.pos, creature.pos) <= hit_r * hit_r:
                    hit_idx = idx
                    break
            if hit_idx is None:
                continue

            creature = creatures[hit_idx]
            hits.append(
                ProjectileHit(
                    type_id=proj.type_id,
                    origin=proj.origin,
                    hit=proj.pos,
                    target=creature.pos,
                ),
            )

            creature = creatures[hit_idx]
            creature.hp -= damage_by_type.get(proj.type_id, 10.0)

            proj.life_timer = 0.25

        return hits



from __future__ import annotations

from .behaviors import PROJECTILE_BEHAVIOR_BY_TYPE_ID, ProjectileBehavior, projectile_behavior_for_type_id
from .collision import _within_native_find_radius
from .primary_rules import PRIMARY_PROJECTILE_RULE_BY_TYPE_ID, PrimaryProjectileRule, primary_rule_for_type_id
from .projectile_pool import PrimaryStepCtx, ProjectilePool, ProjectileUpdateOptions, projectile_collision_profile
from .secondary_pool import SecondaryProjectilePool, SecondarySpawnSpec, SecondaryStepCtx
from .secondary_rules import SECONDARY_RULE_BY_TYPE_ID, secondary_rule_for_type_id

__all__ = [
    "PRIMARY_PROJECTILE_RULE_BY_TYPE_ID",
    "PROJECTILE_BEHAVIOR_BY_TYPE_ID",
    "PrimaryProjectileRule",
    "PrimaryStepCtx",
    "ProjectileBehavior",
    "ProjectilePool",
    "ProjectileUpdateOptions",
    "SECONDARY_RULE_BY_TYPE_ID",
    "SecondarySpawnSpec",
    "SecondaryStepCtx",
    "SecondaryProjectilePool",
    "_within_native_find_radius",
    "projectile_behavior_for_type_id",
    "projectile_collision_profile",
    "primary_rule_for_type_id",
    "secondary_rule_for_type_id",
]



from __future__ import annotations

import hashlib
from pathlib import Path
from typing import TYPE_CHECKING, SupportsInt, cast

import msgspec

from .persistence.save_status import (
    QUEST_PLAY_COUNT,
    UNKNOWN_TAIL_SIZE,
    WEAPON_USAGE_COUNT,
    GameStatus,
    build_status_blob,
    default_status_data,
)
from .replay.types import ReplayStatusSnapshot

if TYPE_CHECKING:
    from .dbg.canonical_channels import SnapshotStatus
    from .net.lockstep_protocol import StatusSnapshot

_ZERO_WEAPON_USAGE_COUNTS: tuple[int, ...] = tuple(0 for _ in range(int(WEAPON_USAGE_COUNT)))
_ZERO_QUEST_PLAY_COUNTS: tuple[int, ...] = tuple(0 for _ in range(int(QUEST_PLAY_COUNT)))
_ZERO_UNKNOWN_TAIL = b"\x00" * int(UNKNOWN_TAIL_SIZE)


class ProgressStatusSnapshot(msgspec.Struct, frozen=True):
    quest_unlock_index: int = 0
    quest_unlock_index_full: int = 0
    weapon_usage_counts: tuple[int, ...] = msgspec.field(default_factory=lambda: _ZERO_WEAPON_USAGE_COUNTS)
    quest_play_counts: tuple[int, ...] = msgspec.field(default_factory=lambda: _ZERO_QUEST_PLAY_COUNTS)
    mode_play_survival: int = 0
    mode_play_rush: int = 0
    mode_play_typo: int = 0
    mode_play_other: int = 0
    game_sequence_id: int = 0
    unknown_tail: bytes = _ZERO_UNKNOWN_TAIL


def _mask_u16(value: object) -> int:
    try:
        return int(cast(SupportsInt, value)) & 0xFFFF
    except (TypeError, ValueError, OverflowError):
        return 0


def _mask_u32(value: object) -> int:
    try:
        return int(cast(SupportsInt, value)) & 0xFFFFFFFF
    except (TypeError, ValueError, OverflowError):
        return 0


def _normalize_u32_seq(raw: object, *, size: int) -> tuple[int, ...]:
    expected = int(size)
    if not isinstance(raw, (list, tuple)):
        return tuple(0 for _ in range(expected))
    values = [int(_mask_u32(value)) for value in list(raw)[:expected]]
    if len(values) < expected:
        values.extend([0] * (expected - len(values)))
    return tuple(values[:expected])


def _normalize_unknown_tail(raw: object) -> bytes:
    if not isinstance(raw, (bytes, bytearray)):
        return _ZERO_UNKNOWN_TAIL
    tail = bytes(raw)
    expected = int(UNKNOWN_TAIL_SIZE)
    if len(tail) < expected:
        tail = tail + (b"\x00" * (expected - len(tail)))
    return tail[:expected]


def _status_blob_data(snapshot: ProgressStatusSnapshot) -> dict[str, object]:
    return {
        "quest_unlock_index": int(snapshot.quest_unlock_index) & 0xFFFF,
        "quest_unlock_index_full": int(snapshot.quest_unlock_index_full) & 0xFFFF,
        "weapon_usage_counts": list(snapshot.weapon_usage_counts),
        "quest_play_counts": list(snapshot.quest_play_counts),
        "mode_play_survival": int(snapshot.mode_play_survival) & 0xFFFFFFFF,
        "mode_play_rush": int(snapshot.mode_play_rush) & 0xFFFFFFFF,
        "mode_play_typo": int(snapshot.mode_play_typo) & 0xFFFFFFFF,
        "mode_play_other": int(snapshot.mode_play_other) & 0xFFFFFFFF,
        "game_sequence_id": int(snapshot.game_sequence_id) & 0xFFFFFFFF,
        "unknown_tail": bytes(snapshot.unknown_tail),
    }


def progress_status_from_game_status(status: GameStatus | None) -> ProgressStatusSnapshot:
    data = {} if status is None else status.data
    return ProgressStatusSnapshot(
        quest_unlock_index=_mask_u16(data.get("quest_unlock_index", 0)),
        quest_unlock_index_full=_mask_u16(data.get("quest_unlock_index_full", 0)),
        weapon_usage_counts=_normalize_u32_seq(data.get("weapon_usage_counts"), size=int(WEAPON_USAGE_COUNT)),
        quest_play_counts=_normalize_u32_seq(data.get("quest_play_counts"), size=int(QUEST_PLAY_COUNT)),
        mode_play_survival=_mask_u32(data.get("mode_play_survival", 0)),
        mode_play_rush=_mask_u32(data.get("mode_play_rush", 0)),
        mode_play_typo=_mask_u32(data.get("mode_play_typo", 0)),
        mode_play_other=_mask_u32(data.get("mode_play_other", 0)),
        game_sequence_id=_mask_u32(data.get("game_sequence_id", 0)),
        unknown_tail=_normalize_unknown_tail(data.get("unknown_tail")),
    )


def game_status_from_progress_status(snapshot: ProgressStatusSnapshot, *, path: Path) -> GameStatus:
    data = default_status_data()
    data.update(_status_blob_data(snapshot))
    return GameStatus(path=path, data=data, dirty=False)


def progress_status_from_lockstep(snapshot: StatusSnapshot | None) -> ProgressStatusSnapshot:
    if snapshot is None:
        return ProgressStatusSnapshot()
    return ProgressStatusSnapshot(
        quest_unlock_index=_mask_u16(snapshot.quest_unlock_index),
        quest_unlock_index_full=_mask_u16(snapshot.quest_unlock_index_full),
        weapon_usage_counts=_normalize_u32_seq(snapshot.weapon_usage_counts, size=int(WEAPON_USAGE_COUNT)),
        quest_play_counts=_normalize_u32_seq(snapshot.quest_play_counts, size=int(QUEST_PLAY_COUNT)),
        mode_play_survival=_mask_u32(snapshot.mode_play_survival),
        mode_play_rush=_mask_u32(snapshot.mode_play_rush),
        mode_play_typo=_mask_u32(snapshot.mode_play_typo),
        mode_play_other=_mask_u32(snapshot.mode_play_other),
        game_sequence_id=_mask_u32(snapshot.game_sequence_id),
        unknown_tail=_normalize_unknown_tail(snapshot.unknown_tail),
    )


def lockstep_status_from_progress(snapshot: ProgressStatusSnapshot) -> StatusSnapshot:
    from .net.lockstep_protocol import StatusSnapshot

    return StatusSnapshot(
        quest_unlock_index=int(snapshot.quest_unlock_index) & 0xFFFF,
        quest_unlock_index_full=int(snapshot.quest_unlock_index_full) & 0xFFFF,
        weapon_usage_counts=list(snapshot.weapon_usage_counts),
        quest_play_counts=list(snapshot.quest_play_counts),
        mode_play_survival=int(snapshot.mode_play_survival) & 0xFFFFFFFF,
        mode_play_rush=int(snapshot.mode_play_rush) & 0xFFFFFFFF,
        mode_play_typo=int(snapshot.mode_play_typo) & 0xFFFFFFFF,
        mode_play_other=int(snapshot.mode_play_other) & 0xFFFFFFFF,
        game_sequence_id=int(snapshot.game_sequence_id) & 0xFFFFFFFF,
        unknown_tail=bytes(snapshot.unknown_tail),
    )


def progress_status_from_replay(snapshot: ReplayStatusSnapshot | None) -> ProgressStatusSnapshot:
    if snapshot is None:
        return ProgressStatusSnapshot()
    return ProgressStatusSnapshot(
        quest_unlock_index=_mask_u16(snapshot.quest_unlock_index),
        quest_unlock_index_full=_mask_u16(snapshot.quest_unlock_index_full),
        weapon_usage_counts=_normalize_u32_seq(snapshot.weapon_usage_counts, size=int(WEAPON_USAGE_COUNT)),
    )


def replay_status_from_progress(snapshot: ProgressStatusSnapshot) -> ReplayStatusSnapshot:
    return ReplayStatusSnapshot(
        quest_unlock_index=int(snapshot.quest_unlock_index) & 0xFFFF,
        quest_unlock_index_full=int(snapshot.quest_unlock_index_full) & 0xFFFF,
        weapon_usage_counts=tuple(snapshot.weapon_usage_counts),
    )


def progress_status_from_debug_snapshot(snapshot: SnapshotStatus | None) -> ProgressStatusSnapshot:
    if snapshot is None:
        return ProgressStatusSnapshot()
    return ProgressStatusSnapshot(
        quest_unlock_index=_mask_u16(snapshot.quest_unlock_index),
        quest_unlock_index_full=_mask_u16(snapshot.quest_unlock_index_full),
        weapon_usage_counts=_normalize_u32_seq(snapshot.weapon_usage_counts, size=int(WEAPON_USAGE_COUNT)),
    )


def debug_snapshot_from_progress_status(snapshot: ProgressStatusSnapshot) -> SnapshotStatus:
    from .dbg.canonical_channels import SnapshotStatus

    return SnapshotStatus(
        quest_unlock_index=int(snapshot.quest_unlock_index) & 0xFFFF,
        quest_unlock_index_full=int(snapshot.quest_unlock_index_full) & 0xFFFF,
        weapon_usage_counts=list(snapshot.weapon_usage_counts),
    )


def hash_progress_status(snapshot: ProgressStatusSnapshot) -> str:
    return hashlib.sha256(build_status_blob(_status_blob_data(snapshot))).hexdigest()



from __future__ import annotations

import hashlib
import platform
import subprocess
from datetime import UTC, datetime
from pathlib import Path
from typing import Literal, cast

import msgspec

from ..replay import load_replay_file
from ..replay.checkpoints import ReplayCheckpoint
from ..replay.types import Replay
from ..sim.driver.playback_driver import PlaybackWalkHooks, build_verify_playback_driver
from ..sim.timing import ftol_ms_i32
from ..sim.world_state import WorldState
from ..status_snapshot import debug_snapshot_from_progress_status, progress_status_from_game_status
from .canonical_channels import (
    BonusEntitySample,
    CreatureEntitySample,
    EntitySamplesSnapshot,
    ProjectileEntitySample,
    RngStreamRow,
    SecondaryProjectileEntitySample,
    SimStateSnapshot,
    SnapshotBonusTimers,
    SnapshotGameplay,
    SnapshotPlayer,
    SnapshotStatus,
    SnapshotVec2,
    SnapshotWeapon,
    bonus_timer_ms,
)
from .checkpoint_codec import checkpoint_to_channel
from .rng import canonical_rng_marks
from .schema import TRACE_FORMAT_VERSION, TRACE_SCHEMA_VERSION, TickRecord, TraceMeta, channel_versions_for
from .trace import TraceError, TraceReader, TraceSummary, write_trace

_REPO_ROOT = Path(__file__).resolve().parents[3]
_ZIG_ROOT = _REPO_ROOT / "crimson-zig"
_ZIG_BIN = _ZIG_ROOT / "zig-out" / "bin" / "crimson-zig"


class _EntityGenerationState(msgspec.Struct):
    generation_by_index: dict[int, int] = msgspec.field(default_factory=dict)
    active_indices: set[int] = msgspec.field(default_factory=set)
    _seen_in_tick: set[int] = msgspec.field(default_factory=set)

    def begin_tick(self) -> None:
        self._seen_in_tick.clear()

    def end_tick(self) -> None:
        self.active_indices = set(self._seen_in_tick)

    def next_generation(self, *, index: int) -> int:
        idx = int(index)
        if idx not in self.generation_by_index:
            self.generation_by_index[idx] = 0
        if idx not in self.active_indices:
            self.generation_by_index[idx] += 1
        self._seen_in_tick.add(idx)
        return int(self.generation_by_index[idx])


def _fingerprint(path: Path) -> dict[str, object]:
    stat = path.stat()
    raw = path.read_bytes()
    return {
        "path": str(path),
        "sha256": hashlib.sha256(raw).hexdigest(),
        "size": stat.st_size,
        "mtime_ns": stat.st_mtime_ns,
    }


def _rng_stream_from_draws(draws: list[tuple[int, int, int]]) -> list[RngStreamRow]:
    rows: list[RngStreamRow] = []
    for index, row in enumerate(draws):
        state_before_u32, value_15, state_after_u32 = row
        rows.append(
            RngStreamRow(
                tick_call_index=int(index) + 1,
                value_15=int(value_15),
                state_before_u32=int(state_before_u32),
                state_after_u32=int(state_after_u32),
                caller_static=None,
                branch_id=None,
            ),
        )
    return rows


def _entity_samples_for_world(
    world: WorldState,
    *,
    creature_state: _EntityGenerationState,
    projectile_state: _EntityGenerationState,
    secondary_state: _EntityGenerationState,
    bonus_state: _EntityGenerationState,
) -> EntitySamplesSnapshot:
    creature_state.begin_tick()
    projectile_state.begin_tick()
    secondary_state.begin_tick()
    bonus_state.begin_tick()

    creatures: list[CreatureEntitySample] = []
    for index, creature in enumerate(world.creatures.entries):
        if not creature.active:
            continue
        generation = creature_state.next_generation(index=index)
        creatures.append(
            CreatureEntitySample(
                uid=int(index),
                generation=generation,
                pool_kind="creature",
                index=index,
                active=True,
                type_id=int(creature.type_id),
                hp=float(creature.hp),
                pos=SnapshotVec2(x=float(creature.pos.x), y=float(creature.pos.y)),
                flags=int(creature.flags),
                ai_mode=int(creature.ai_mode),
                link_index=int(creature.link_index),
                heading=float(creature.heading),
                target_heading=float(creature.target_heading),
                orbit_angle=float(creature.orbit_angle),
                orbit_radius=float(creature.orbit_radius),
                lifecycle_stage=float(creature.lifecycle_stage),
            ),
        )

    projectiles: list[ProjectileEntitySample] = []
    for index, projectile in enumerate(world.state.projectiles.entries):
        if not projectile.active:
            continue
        generation = projectile_state.next_generation(index=index)
        projectiles.append(
            ProjectileEntitySample(
                uid=int(index),
                generation=generation,
                pool_kind="projectile",
                index=index,
                active=True,
                type_id=int(projectile.type_id),
                angle=float(projectile.angle),
                pos=SnapshotVec2(x=float(projectile.pos.x), y=float(projectile.pos.y)),
                vel=SnapshotVec2(x=float(projectile.vel.x), y=float(projectile.vel.y)),
                life_timer=float(projectile.life_timer),
                speed_scale=float(projectile.speed_scale),
                damage_pool=float(projectile.damage_pool),
                hit_radius=float(projectile.hit_radius),
                travel_budget=float(projectile.travel_budget),
                owner=projectile.owner,
            ),
        )

    secondary_projectiles: list[SecondaryProjectileEntitySample] = []
    for index, projectile in enumerate(world.state.secondary_projectiles.entries):
        if not projectile.active:
            continue
        generation = secondary_state.next_generation(index=index)
        secondary_projectiles.append(
            SecondaryProjectileEntitySample(
                uid=int(index),
                generation=generation,
                pool_kind="secondary_projectile",
                index=index,
                active=True,
                type_id=int(projectile.type_id),
                angle=float(projectile.angle),
                pos=SnapshotVec2(x=float(projectile.pos.x), y=float(projectile.pos.y)),
                vel=SnapshotVec2(x=float(projectile.vel.x), y=float(projectile.vel.y)),
                speed=float(projectile.speed),
                trail_timer=float(projectile.trail_timer),
                owner=projectile.owner,
                target_id=int(projectile.target_id),
            ),
        )

    bonuses: list[BonusEntitySample] = []
    for index, bonus in enumerate(world.state.bonus_pool.entries):
        if int(bonus.bonus_id) == 0:
            continue
        generation = bonus_state.next_generation(index=index)
        bonuses.append(
            BonusEntitySample(
                uid=int(index),
                generation=generation,
                pool_kind="bonus",
                index=index,
                active=True,
                bonus_id=int(bonus.bonus_id),
                picked=bool(bonus.picked),
                time_left=float(bonus.time_left),
                time_max=float(bonus.time_max),
                pos=SnapshotVec2(x=float(bonus.pos.x), y=float(bonus.pos.y)),
                amount=int(bonus.amount),
            ),
        )

    creature_state.end_tick()
    projectile_state.end_tick()
    secondary_state.end_tick()
    bonus_state.end_tick()

    return EntitySamplesSnapshot(
        creatures=creatures,
        projectiles=projectiles,
        secondary_projectiles=secondary_projectiles,
        bonuses=bonuses,
    )


def _status_snapshot_from_world(world: WorldState) -> SnapshotStatus:
    gameplay = world.state
    if gameplay.status is None:
        raise ValueError("gameplay status missing while recording trace")
    return debug_snapshot_from_progress_status(progress_status_from_game_status(gameplay.status))


def _sim_state_from_world(world: WorldState, *, replay: Replay) -> SimStateSnapshot:
    gameplay = world.state
    players: list[SnapshotPlayer] = []
    for player in world.players:
        players.append(
            SnapshotPlayer(
                index=int(player.index),
                pos=SnapshotVec2(x=float(player.pos.x), y=float(player.pos.y)),
                health=float(player.health),
                weapon=SnapshotWeapon(
                    weapon_id=int(player.weapon.weapon_id),
                    ammo=float(player.weapon.ammo),
                    clip_size=int(player.weapon.clip_size),
                    reload_active=bool(player.weapon.reload_active),
                    reload_timer=float(player.weapon.reload_timer),
                    reload_timer_max=float(player.weapon.reload_timer_max),
                    shot_cooldown=float(player.weapon.shot_cooldown),
                ),
                experience=int(player.experience),
                level=int(player.level),
            ),
        )
    return SimStateSnapshot(
        gameplay=SnapshotGameplay(
            mode_id=int(replay.header.game_mode_id),
            quest_stage_major=int(gameplay.quest_stage_major),
            quest_stage_minor=int(gameplay.quest_stage_minor),
            perk_pending_count=int(gameplay.perk_selection.pending_count),
            perk_choices_dirty=bool(gameplay.perk_selection.choices_dirty),
            bonus_timers=SnapshotBonusTimers(
                weapon_power_up_ms=bonus_timer_ms(float(gameplay.bonuses.weapon_power_up)),
                reflex_boost_ms=bonus_timer_ms(float(gameplay.bonuses.reflex_boost)),
                energizer_ms=bonus_timer_ms(float(gameplay.bonuses.energizer)),
                double_experience_ms=bonus_timer_ms(float(gameplay.bonuses.double_experience)),
                freeze_ms=bonus_timer_ms(float(gameplay.bonuses.freeze)),
            ),
            status=_status_snapshot_from_world(world),
        ),
        players=players,
    )


def _build_replay_fingerprint(*, replay_path: Path, replay: Replay) -> dict[str, object]:
    replay_fingerprint = _fingerprint(replay_path)
    replay_fingerprint["tick_rate"] = replay.header.tick_rate
    replay_fingerprint["seed"] = replay.header.seed
    replay_fingerprint["mode_id"] = replay.header.game_mode_id
    replay_fingerprint["quest_level"] = str(replay.header.quest_level)
    return replay_fingerprint


def _build_trace_meta(
    *,
    replay_path: Path,
    replay: Replay,
    tick_rows: list[TickRecord],
    channels_seen: set[str],
    impl: Literal["python", "zig"],
    config_extra: dict[str, object] | None = None,
) -> TraceMeta:
    tick_start = min((row.tick_index for row in tick_rows), default=-1)
    tick_end = max((row.tick_index for row in tick_rows), default=-1)
    replay_fingerprint = _build_replay_fingerprint(replay_path=replay_path, replay=replay)
    channels_sorted = sorted(channels_seen)
    config = {
        "impl": str(impl),
    }
    if config_extra is not None:
        config.update(config_extra)
    return TraceMeta(
        trace_format_version=TRACE_FORMAT_VERSION,
        trace_schema_version=TRACE_SCHEMA_VERSION,
        created_utc=datetime.now(tz=UTC).isoformat(),
        producer={
            "impl": str(impl),
            "impl_version": "",
            "platform": str(platform.system()),
            "arch": str(platform.machine()),
        },
        source=replay_fingerprint,
        channels=channels_sorted,
        channel_versions=channel_versions_for(channels_sorted),
        tick_range={
            "start_tick": tick_start,
            "end_tick": tick_end,
            "tick_count": len(tick_rows),
        },
        config=config,
    )


def _record_replay_to_trace_python(
    *,
    replay_path: Path,
    out_path: Path,
    chunk_ticks: int,
) -> TraceSummary:
    replay = load_replay_file(replay_path)

    replay_tick_count = len(replay.ticks)
    checkpoint_ticks = set(range(replay_tick_count))
    checkpoints: list[ReplayCheckpoint] = []

    entity_samples_by_tick: dict[int, EntitySamplesSnapshot] = {}
    sim_state_by_tick: dict[int, SimStateSnapshot] = {}
    rng_stream_by_tick: dict[int, list[RngStreamRow]] = {}
    creature_state = _EntityGenerationState()
    projectile_state = _EntityGenerationState()
    secondary_state = _EntityGenerationState()
    bonus_state = _EntityGenerationState()

    def _tick_observer(tick_index: int, world: WorldState) -> None:
        entity_samples_by_tick[tick_index] = _entity_samples_for_world(
            world,
            creature_state=creature_state,
            projectile_state=projectile_state,
            secondary_state=secondary_state,
            bonus_state=bonus_state,
        )
        sim_state_by_tick[tick_index] = _sim_state_from_world(world, replay=replay)

    def _tick_rng_trace_observer(tick_index: int, draws: list[tuple[int, int, int]]) -> None:
        rng_stream_by_tick[int(tick_index)] = _rng_stream_from_draws(draws)

    driver = build_verify_playback_driver(
        replay,
        max_ticks=None,
        trace_rng=True,
    )

    def _after_tick(tick_result, world: WorldState) -> None:
        tick_index = int(tick_result.source_tick.tick_index)
        if tick_index in checkpoint_ticks:
            checkpoints.append(driver.build_checkpoint(tick_result=tick_result))
        _tick_observer(tick_index, world)

    def _on_rng_trace(tick_result, draws: tuple[tuple[int, int, int], ...]) -> None:
        _tick_rng_trace_observer(
            int(tick_result.source_tick.tick_index),
            list(draws),
        )

    driver.run(
        hooks=PlaybackWalkHooks(
            after_tick=_after_tick,
            on_rng_trace=_on_rng_trace,
        ),
    )

    tick_rows: list[TickRecord] = []
    channels_seen: set[str] = set()
    replay_dt_rows = [ftol_ms_i32(tick.dt) for tick in replay.ticks]
    for checkpoint in sorted(checkpoints, key=lambda row: row.tick_index):
        tick_index = int(checkpoint.tick_index)
        if tick_index not in rng_stream_by_tick:
            raise ValueError(f"missing runtime rng stream for tick {tick_index}")
        if tick_index not in entity_samples_by_tick:
            raise ValueError(f"missing entity_samples snapshot for tick {tick_index}")
        if tick_index not in sim_state_by_tick:
            raise ValueError(f"missing sim_state snapshot for tick {tick_index}")

        entity_samples_obj = entity_samples_by_tick[tick_index]
        sim_state_obj = sim_state_by_tick[tick_index]
        rng_stream = list(rng_stream_by_tick[tick_index])
        trace_rng_marks = canonical_rng_marks(
            rng_state=int(checkpoint.rng_state),
            rng_stream=rng_stream,
        )
        trace_checkpoint = msgspec.structs.replace(
            checkpoint,
            rng_marks=dict(trace_rng_marks),
        )

        channels: dict[str, object] = {
            "checkpoint": checkpoint_to_channel(trace_checkpoint),
            "sim_state": cast("dict[str, object]", msgspec.to_builtins(sim_state_obj)),
            "entity_samples": cast("dict[str, object]", msgspec.to_builtins(entity_samples_obj)),
            "rng_marks": dict(trace_rng_marks),
            "rng_stream": msgspec.to_builtins(rng_stream),
            "timing_samples": [],
        }

        if not (0 <= tick_index < len(replay_dt_rows)):
            raise ValueError(f"missing replay dt_ms_i32 row for tick {tick_index}")
        tick_dt_ms_i32 = int(replay_dt_rows[tick_index])
        if tick_dt_ms_i32 < 0:
            raise ValueError(f"invalid replay dt_ms_i32 at tick {tick_index}: {tick_dt_ms_i32}")

        channels_seen.update(channels.keys())
        tick_rows.append(
            TickRecord(
                tick_index=tick_index,
                elapsed_ms=int(checkpoint.elapsed_ms),
                dt_ms_i32=tick_dt_ms_i32,
                mode_id=int(replay.header.game_mode_id),
                phase_markers=[],
                channels=channels,
            ),
        )

    meta = _build_trace_meta(
        replay_path=replay_path,
        replay=replay,
        tick_rows=tick_rows,
        channels_seen=channels_seen,
        impl="python",
        config_extra=None,
    )
    return write_trace(
        out_path,
        meta=meta,
        ticks=tick_rows,
        chunk_ticks=max(1, chunk_ticks),
    )


def _command_detail(run: subprocess.CompletedProcess[str]) -> str:
    stderr = str(run.stderr).strip()
    stdout = str(run.stdout).strip()
    if stderr:
        return stderr
    if stdout:
        return stdout
    return "(no command output)"


def _run_process(command: list[str], *, cwd: Path) -> subprocess.CompletedProcess[str]:
    try:
        return subprocess.run(
            command,
            cwd=cwd,
            check=False,
            text=True,
            capture_output=True,
        )
    except OSError as exc:
        joined = " ".join(command)
        raise ValueError(f"failed to run command: {joined}: {exc}") from exc


def _record_replay_to_trace_zig(
    *,
    replay_path: Path,
    out_path: Path,
    chunk_ticks: int,
) -> tuple[TraceSummary, list[str]]:
    replay = load_replay_file(replay_path)
    if int(replay.header.player_count) != 1:
        raise ValueError(
            "zig dbg record currently supports only single-player replays; "
            f"got player_count={int(replay.header.player_count)}",
        )

    build_run = _run_process(["zig", "build"], cwd=_ZIG_ROOT)
    if int(build_run.returncode) != 0:
        raise ValueError(
            "zig build failed: "
            f"exit={int(build_run.returncode)} detail={_command_detail(build_run)}",
        )

    warnings: list[str] = []
    verify_cmd = [
        str(_ZIG_BIN),
        "replay",
        "verify",
        str(replay_path),
        "--debug-trace-cdt",
        str(out_path),
        "--debug-trace-cdt-chunk-ticks",
        str(max(1, int(chunk_ticks))),
        "--format",
        "json",
    ]
    verify_run = _run_process(verify_cmd, cwd=_REPO_ROOT)
    if not out_path.is_file():
        raise ValueError(
            "zig trace generation failed: replay verify did not produce CDT output; "
            f"exit={int(verify_run.returncode)} detail={_command_detail(verify_run)}",
        )
    if int(verify_run.returncode) != 0:
        warning = (
            f"warning: zig replay verify exited {int(verify_run.returncode)}; "
            "continuing with emitted trace"
        )
        stderr = str(verify_run.stderr).strip()
        if stderr:
            warning = f"{warning}: {stderr.splitlines()[0]}"
        warnings.append(warning)

    try:
        with TraceReader(out_path) as trace:
            summary = TraceSummary(meta=trace.meta, footer=trace.footer)
    except TraceError as exc:
        raise ValueError(f"zig trace validation failed: {exc}") from exc
    return summary, warnings


def record_replay_to_trace(
    *,
    replay_path: Path,
    out_path: Path,
    impl: Literal["python", "zig"] = "python",
    chunk_ticks: int = 256,
    warnings_out: list[str] | None = None,
) -> TraceSummary:
    replay_path = Path(replay_path)
    out_path = Path(out_path)
    if warnings_out is None:
        warnings_out = []
    if str(impl) == "python":
        summary = _record_replay_to_trace_python(
            replay_path=replay_path,
            out_path=out_path,
            chunk_ticks=chunk_ticks,
        )
        return summary
    if str(impl) == "zig":
        summary, warnings = _record_replay_to_trace_zig(
            replay_path=replay_path,
            out_path=out_path,
            chunk_ticks=chunk_ticks,
        )
        warnings_out.extend(warnings)
        return summary
    raise ValueError(f"unsupported dbg record impl: {impl!r}")



from __future__ import annotations

import io
from collections.abc import Sequence
from pathlib import Path
from typing import Protocol, cast

import msgspec
import zstandard as zstd

from grim.geom import Vec2

from ..bonuses import BonusId
from ..owner_ref import OwnerRef
from ..sim.state_types import PlayerState
from ..sim.timing import ftol_ms_i32
from ..sim.world_state import WorldState
from ..weapons import WeaponId


class _DeathLike(Protocol):
    index: int
    type_id: int
    reward_value: float
    xp_awarded: int
    owner: OwnerRef


class _EventsLike(Protocol):
    hits: Sequence[object]
    pickups: list[object]
    sfx: list[str]

FORMAT_VERSION = 3
DEFAULT_CHECKPOINT_SAMPLE_RATE = 1
_ZSTD_MAGIC = b"\x28\xb5\x2f\xfd"
_DEFAULT_MAX_CHECKPOINTS_PAYLOAD_BYTES = 256 * 1024 * 1024
_CHECKPOINTS_ZSTD_LEVEL = 19


class ReplayCheckpointsError(ValueError):
    pass


class ReplayPlayerCheckpoint(msgspec.Struct, frozen=True):
    pos: Vec2
    health: float
    weapon_id: WeaponId
    ammo: float
    experience: int
    level: int

class ReplayCheckpoint(msgspec.Struct, frozen=True):
    tick_index: int
    rng_state: int
    elapsed_ms: int
    score_xp: int
    kills: int
    creature_count: int
    perk_pending: int
    players: list[ReplayPlayerCheckpoint]
    bonus_timers: dict[str, int]
    rng_marks: dict[str, int] = msgspec.field(default_factory=dict)
    deaths: list["ReplayDeathLedgerEntry"] = msgspec.field(default_factory=list)
    perk: "ReplayPerkSnapshot" = msgspec.field(default_factory=lambda: ReplayPerkSnapshot())
    events: "ReplayEventSummary" = msgspec.field(default_factory=lambda: ReplayEventSummary())


class ReplayDeathLedgerEntry(msgspec.Struct, frozen=True):
    creature_index: int
    type_id: int
    reward_value: float
    xp_awarded: int
    owner: OwnerRef


class ReplayPerkSnapshot(msgspec.Struct, frozen=True):
    pending_count: int = 0
    choices_dirty: bool = False
    choices: list[int] = msgspec.field(default_factory=list)
    player_nonzero_counts: list[list[list[int]]] = msgspec.field(default_factory=list)


class ReplayEventSummary(msgspec.Struct, frozen=True):
    # Legacy sidecars may omit this block; -1 marks "unknown/not recorded".
    hit_count: int = 0
    pickup_count: int = 0
    sfx_count: int = 0
    sfx_head: list[str] = msgspec.field(default_factory=list)


class ReplayCheckpoints(msgspec.Struct, frozen=True):
    version: int
    sample_rate: int
    checkpoints: list[ReplayCheckpoint] = msgspec.field(default_factory=list)

_CHECKPOINTS_ENCODER = msgspec.msgpack.Encoder()
_CHECKPOINTS_DECODER = msgspec.msgpack.Decoder(type=ReplayCheckpoints)


def default_checkpoints_path(replay_path: Path) -> Path:
    replay_path = Path(replay_path)
    return replay_path.with_name(f"{replay_path.name}.chk")


def _is_zstd(data: bytes) -> bool:
    return bytes(data).startswith(_ZSTD_MAGIC)


def _decompress_zstd_checkpoints(data: bytes, *, max_output_bytes: int) -> bytes:
    try:
        with zstd.ZstdDecompressor().stream_reader(io.BytesIO(data)) as stream:
            payload = stream.read(int(max_output_bytes) + 1)
    except zstd.ZstdError as exc:
        raise ReplayCheckpointsError("invalid checkpoints zstd payload") from exc
    if len(payload) > int(max_output_bytes):
        raise ReplayCheckpointsError(
            f"checkpoints payload too large after zstd decompression (> {int(max_output_bytes)} bytes)",
        )
    return payload


def _bonus_timer_ms(value: float) -> int:
    # Keep checkpoint values compact/stable: ms resolution is enough for divergence detection.
    ms = int(ftol_ms_i32(float(value)))
    if ms < 0:
        return 0
    return ms


def build_checkpoint(
    *,
    tick_index: int,
    world: WorldState,
    elapsed_ms: float,
    creature_count_override: int | None = None,
    rng_marks: dict[str, int] | None = None,
    deaths: Sequence[object] | None = None,
    events: object | None = None,
) -> ReplayCheckpoint:
    state = world.state
    players: list[PlayerState] = list(world.players)
    score_xp = sum(int(player.experience) for player in players)
    kills = int(world.creatures.kill_count)
    if creature_count_override is None:
        creature_count = sum(1 for creature in world.creatures.entries if creature.active)
    else:
        creature_count = int(creature_count_override)

    player_ckpts: list[ReplayPlayerCheckpoint] = []
    for player in players:
        player_ckpts.append(
            ReplayPlayerCheckpoint(
                pos=Vec2(round(player.pos.x, 4), round(player.pos.y, 4)),
                health=round(player.health, 4),
                weapon_id=WeaponId(player.weapon.weapon_id),
                ammo=round(player.weapon.ammo, 4),
                experience=int(player.experience),
                level=int(player.level),
            ),
        )

    bonus_timers = {
        str(BonusId.WEAPON_POWER_UP): _bonus_timer_ms(state.bonuses.weapon_power_up),
        str(BonusId.REFLEX_BOOST): _bonus_timer_ms(state.bonuses.reflex_boost),
        str(BonusId.ENERGIZER): _bonus_timer_ms(state.bonuses.energizer),
        str(BonusId.DOUBLE_EXPERIENCE): _bonus_timer_ms(state.bonuses.double_experience),
        str(BonusId.FREEZE): _bonus_timer_ms(state.bonuses.freeze),
    }

    perk_counts: list[list[list[int]]] = []
    for player in players:
        nonzero: list[list[int]] = []
        for perk_id, count in enumerate(player.perk_counts):
            count_i = int(count)
            if count_i != 0:
                nonzero.append([int(perk_id), count_i])
        perk_counts.append(nonzero)

    perk_snapshot = ReplayPerkSnapshot(
        pending_count=int(state.perk_selection.pending_count),
        choices_dirty=bool(state.perk_selection.choices_dirty),
        choices=[int(perk_id) for perk_id in state.perk_selection.choices],
        player_nonzero_counts=perk_counts,
    )

    death_entries: list[ReplayDeathLedgerEntry] = []
    for death in deaths or ():
        death_view = cast(_DeathLike, death)
        death_entries.append(
            ReplayDeathLedgerEntry(
                creature_index=int(death_view.index),
                type_id=int(death_view.type_id),
                reward_value=float(death_view.reward_value),
                xp_awarded=int(death_view.xp_awarded),
                owner=death_view.owner,
            ),
        )

    marks: dict[str, int] = {}
    if rng_marks:
        for key, value in rng_marks.items():
            marks[str(key)] = int(value)

    events_view = cast(_EventsLike | None, events)
    hits = list(events_view.hits) if events_view is not None else []
    pickups = list(events_view.pickups) if events_view is not None else []
    sfx = list(events_view.sfx) if events_view is not None else []
    event_summary = ReplayEventSummary(
        hit_count=len(hits),
        pickup_count=len(pickups),
        sfx_count=len(sfx),
        sfx_head=[str(key) for key in sfx[:4]],
    )

    return ReplayCheckpoint(
        tick_index=int(tick_index),
        rng_state=int(state.rng.state),
        elapsed_ms=int(round(elapsed_ms)),
        score_xp=int(score_xp),
        kills=int(kills),
        creature_count=int(creature_count),
        perk_pending=int(state.perk_selection.pending_count),
        players=player_ckpts,
        bonus_timers=bonus_timers,
        rng_marks=marks,
        deaths=death_entries,
        perk=perk_snapshot,
        events=event_summary,
    )


def dump_checkpoints(checkpoints: ReplayCheckpoints) -> bytes:
    raw = _CHECKPOINTS_ENCODER.encode(checkpoints)
    return zstd.ZstdCompressor(level=_CHECKPOINTS_ZSTD_LEVEL).compress(raw)


def load_checkpoints(data: bytes) -> ReplayCheckpoints:
    max_payload_bytes = int(_DEFAULT_MAX_CHECKPOINTS_PAYLOAD_BYTES)
    payload = bytes(data)
    if _is_zstd(payload):
        payload = _decompress_zstd_checkpoints(payload, max_output_bytes=max_payload_bytes)
    if len(payload) > int(max_payload_bytes):
        raise ReplayCheckpointsError(f"checkpoints payload too large (> {int(max_payload_bytes)} bytes)")
    try:
        decoded = _CHECKPOINTS_DECODER.decode(payload)
    except (msgspec.DecodeError, msgspec.ValidationError) as exc:
        raise ReplayCheckpointsError("invalid checkpoints msgpack payload") from exc

    if int(decoded.version) != FORMAT_VERSION:
        raise ReplayCheckpointsError(f"unsupported checkpoints version: {int(decoded.version)}")
    return decoded


def dump_checkpoints_file(path: Path, checkpoints: ReplayCheckpoints) -> None:
    Path(path).write_bytes(dump_checkpoints(checkpoints))


def load_checkpoints_file(path: Path) -> ReplayCheckpoints:
    return load_checkpoints(Path(path).read_bytes())



from __future__ import annotations

import math
from collections.abc import Callable
from pathlib import Path

import msgspec

from grim.assets import TextureLoader
from grim.config import CrimsonConfig
from grim.fonts.small import SmallFontData, draw_small_text, load_small_font, measure_small_text_width
from grim.geom import Rect, Vec2
from grim.raylib_api import rl

from ..game_modes import GameMode
from ..persistence.highscores import (
    NAME_MAX_EDIT,
    TABLE_MAX,
    HighScoreRecord,
    rank_index,
    read_highscore_table,
    scores_path_for_config,
    upsert_highscore_record,
)
from ..weapons import WEAPON_BY_ID, WeaponId, weapon_display_name
from .cursor import draw_menu_cursor
from .formatting import format_ordinal, format_time_mm_ss
from .hud import HudAssets
from .layout import menu_widescreen_y_shift, ui_scale
from .menu_panel import draw_classic_menu_panel
from .perk_menu import (
    PerkMenuAssets,
    UiButtonState,
    button_draw,
    button_update,
    button_width,
    draw_ui_text,
    load_perk_menu_assets,
)
from .text_input import flush_text_input_events, gameplay_controls_held, poll_text_input

GAME_OVER_PANEL_X = -45.0
# `ui_menu_layout_init` sets game-over panel pos to (-45, 110):
#   _DAT_0048cc60 = 0xc2340000 (-45.0)
#   _DAT_0048cc64 = 0x42dc0000 (110.0)
GAME_OVER_PANEL_Y = 110.0
# `DAT_0048cc48` is cloned from the 3-slice menu panel layout (`ui_menu_item_element._pad4+0xac`)
# in `ui_menu_layout_init`; trace confirms a 510x378 bbox for both phase 0 and phase 1.
GAME_OVER_PANEL_W = 510.0
GAME_OVER_PANEL_H = 378.0

# Measured from ui_render_trace at 1024x768 (stable timeline):
# panel top-left is (pos_x + 21, pos_y - 81) and size is 510x254, plus a shadow pass at +7,+7.
GAME_OVER_PANEL_OFFSET_X = 21.0
GAME_OVER_PANEL_OFFSET_Y = -81.0

TEXTURE_TOP_BANNER_W = 256.0
TEXTURE_TOP_BANNER_H = 64.0

# `game_over_screen_update` (0x0040ffc0) computes banner/content X from:
#   local_10 = quad0_x0 + pos_x + 180.0
#   local_18 = offset_x + local_10 + 44.0 - 10.0
# so banner/content anchor is +214 from the panel-left edge in steady state.
GAME_OVER_BANNER_X_OFFSET = 214.0

INPUT_BOX_W = 166.0  # `_DAT_0048259c = 0xa6` before `ui_text_input_update`
INPUT_BOX_H = 18.0

PANEL_SLIDE_DURATION_MS = 250.0

COLOR_TEXT = rl.Color(255, 255, 255, 255)
COLOR_TEXT_MUTED = rl.Color(255, 255, 255, int(255 * 0.8))
COLOR_SCORE_LABEL = rl.Color(230, 230, 230, 255)
COLOR_SCORE_VALUE = rl.Color(230, 230, 255, 255)


def _weapon_icon_src(texture: rl.Texture, weapon_id_native: int) -> rl.Rectangle | None:
    weapon_id = int(weapon_id_native)
    icon_index = WEAPON_BY_ID[WeaponId(weapon_id)].icon_index
    if icon_index < 0 or icon_index > 31:
        return None
    grid = 8
    cell_w = float(texture.width) / grid
    cell_h = float(texture.height) / grid
    frame = int(icon_index) * 2
    col = frame % grid
    row = frame // grid
    return rl.Rectangle(float(col * cell_w), float(row * cell_h), float(cell_w * 2), float(cell_h))


class GameOverAssets(msgspec.Struct):
    menu_panel: rl.Texture | None
    text_reaper: rl.Texture | None
    text_well_done: rl.Texture | None
    particles: rl.Texture | None
    perk_menu_assets: PerkMenuAssets


class _GameOverPanelLayout(msgspec.Struct, frozen=True):
    panel: Rect
    top_left: Vec2


def load_game_over_assets(assets_root: Path) -> GameOverAssets:
    perk_menu_assets = load_perk_menu_assets(assets_root)
    loader = TextureLoader.from_assets_root(assets_root)
    menu_panel = loader.get(name="ui_menuPanel", paq_rel="ui/ui_menuPanel.jaz")
    text_reaper = loader.get(name="ui_textReaper", paq_rel="ui/ui_textReaper.jaz")
    text_well_done = loader.get(
        name="ui_textWellDone",
        paq_rel="ui/ui_textWellDone.jaz",
    )
    particles = loader.get(name="particles", paq_rel="game/particles.jaz")
    return GameOverAssets(
        menu_panel=menu_panel,
        text_reaper=text_reaper,
        text_well_done=text_well_done,
        particles=particles,
        perk_menu_assets=perk_menu_assets,
    )


def _draw_texture_centered(tex: rl.Texture, pos: Vec2, w: float, h: float, alpha: float) -> None:
    src = rl.Rectangle(0.0, 0.0, float(tex.width), float(tex.height))
    dst = rl.Rectangle(pos.x, pos.y, float(w), float(h))
    tint = rl.Color(255, 255, 255, int(255 * max(0.0, min(1.0, alpha))))
    rl.draw_texture_pro(tex, src, dst, rl.Vector2(0.0, 0.0), 0.0, tint)


def _ease_out_cubic(t: float) -> float:
    t = max(0.0, min(1.0, float(t)))
    return 1.0 - (1.0 - t) ** 3


class GameOverUi(msgspec.Struct):
    assets_root: Path
    base_dir: Path

    config: CrimsonConfig
    preserve_bugs: bool = False

    assets: GameOverAssets | None = None
    font: SmallFontData | None = None

    input_text: str = ""
    input_caret: int = 0
    phase: int = -1  # -1 init, 0 name entry (if qualifies), 1 results/buttons
    rank: int = TABLE_MAX
    _candidate_record: HighScoreRecord | None = None
    _saved: bool = False
    _dt: float = 0.0

    _hover_weapon: float = 0.0
    _hover_time: float = 0.0
    _hover_hit_ratio: float = 0.0
    _intro_ms: float = 0.0
    _cursor_pulse_time: float = 0.0
    _panel_open_sfx_played: bool = False
    _closing: bool = False
    _close_action: str | None = None

    # Buttons (rendered via existing ui_button implementation)
    _ok_button: UiButtonState = msgspec.field(default_factory=lambda: UiButtonState("OK", force_wide=False))
    _play_again_button: UiButtonState = msgspec.field(default_factory=lambda: UiButtonState("Play Again", force_wide=True))
    _high_scores_button: UiButtonState = msgspec.field(default_factory=lambda: UiButtonState("High scores", force_wide=True))
    _main_menu_button: UiButtonState = msgspec.field(default_factory=lambda: UiButtonState("Main Menu", force_wide=True))

    _consume_enter: bool = False
    _defer_name_input_until_controls_released: bool = False

    def open(self) -> None:
        self.close()
        self.font = load_small_font(self.assets_root)
        self.assets = load_game_over_assets(self.assets_root)
        self.phase = -1
        self.rank = TABLE_MAX
        self._candidate_record = None
        self._saved = False
        self._dt = 0.0
        self._hover_weapon = 0.0
        self._hover_time = 0.0
        self._hover_hit_ratio = 0.0
        self._intro_ms = 0.0
        self._cursor_pulse_time = 0.0
        self._panel_open_sfx_played = False
        self._closing = False
        self._close_action = None
        self.input_text = ""
        self.input_caret = 0
        self._consume_enter = True
        self._defer_name_input_until_controls_released = False

    def close(self) -> None:
        if self.assets is not None:
            self.assets = None
        if self.font is not None:
            rl.unload_texture(self.font.texture)
            self.font = None

    def consume_enter(self) -> bool:
        if self._consume_enter:
            self._consume_enter = False
            return True
        return False

    def world_entity_alpha(self) -> float:
        if not self._closing:
            return 1.0
        if PANEL_SLIDE_DURATION_MS <= 1e-6:
            return 0.0
        alpha = float(self._intro_ms) / float(PANEL_SLIDE_DURATION_MS)
        if alpha < 0.0:
            return 0.0
        if alpha > 1.0:
            return 1.0
        return alpha

    def _text_width(self, text: str, scale: float) -> float:
        if self.font is None:
            return float(rl.measure_text(text, int(20 * scale)))
        return float(measure_small_text_width(self.font, text, scale))

    def _draw_small(self, text: str, pos: Vec2, scale: float, color: rl.Color) -> None:
        if self.font is not None:
            draw_small_text(self.font, text, pos, scale, color)
        else:
            rl.draw_text(text, int(pos.x), int(pos.y), int(20 * scale), color)

    def _panel_layout(self, *, screen_w: float, scale: float) -> _GameOverPanelLayout:
        # Keep consistent with the main menu panel offsets.
        t = self._intro_ms / PANEL_SLIDE_DURATION_MS if PANEL_SLIDE_DURATION_MS > 1e-6 else 1.0
        eased = _ease_out_cubic(t)
        panel_slide_x = -GAME_OVER_PANEL_W * (1.0 - eased)

        panel_pos = Vec2((GAME_OVER_PANEL_X + panel_slide_x) * scale, 0.0)
        layout_w = screen_w / scale if scale else screen_w
        widescreen_shift_y = menu_widescreen_y_shift(layout_w)
        panel_pos = Vec2(panel_pos.x, (GAME_OVER_PANEL_Y + widescreen_shift_y) * scale)
        panel_origin = Vec2(-(GAME_OVER_PANEL_OFFSET_X * scale), -(GAME_OVER_PANEL_OFFSET_Y * scale))
        top_left = panel_pos - panel_origin
        panel = Rect.from_top_left(top_left, GAME_OVER_PANEL_W * scale, GAME_OVER_PANEL_H * scale)
        return _GameOverPanelLayout(panel=panel, top_left=top_left)

    def _begin_close_transition(self, action: str) -> None:
        if self._closing:
            return
        self._closing = True
        self._close_action = action

    def update(
        self,
        dt: float,
        *,
        record: HighScoreRecord,
        player_name_default: str,
        play_sfx: Callable[[str], None] | None = None,
        rand: Callable[[], int] | None = None,
        mouse: rl.Vector2 | None = None,
    ) -> str | None:
        self._dt = float(min(dt, 0.1))
        dt_ms = self._dt * 1000.0
        self._cursor_pulse_time += self._dt * 1.1
        if mouse is None:
            mouse = rl.get_mouse_position()
        if rand is None:

            def rand() -> int:
                return 0

        if self.assets is None:
            return None

        if self._closing:
            self._intro_ms = max(0.0, float(self._intro_ms) - dt_ms)
            if self._intro_ms <= 1e-3 and self._close_action is not None:
                action = self._close_action
                self._close_action = None
                self._closing = False
                return action
            return None

        self._intro_ms = min(PANEL_SLIDE_DURATION_MS, self._intro_ms + dt_ms)
        if (
            (not self._panel_open_sfx_played)
            and play_sfx is not None
            and self._intro_ms >= PANEL_SLIDE_DURATION_MS - 1e-3
        ):
            play_sfx("sfx_ui_panelclick")
            self._panel_open_sfx_played = True
        if self._consume_enter:
            self._consume_enter = False
            rl.is_key_pressed(rl.KeyboardKey.KEY_ENTER)
        if self.phase == -1:
            # If in the top 100, prompt for a name. Otherwise show score-too-low message and buttons.
            try:
                game_mode_id = GameMode(self.config.game_mode)
            except ValueError:
                game_mode_id = GameMode.DEMO
            candidate = record.copy()
            candidate.game_mode_id = game_mode_id
            self._candidate_record = candidate

            path = scores_path_for_config(self.base_dir, self.config)
            records = read_highscore_table(path, game_mode_id=game_mode_id)
            idx = rank_index(records, candidate)
            self.rank = int(idx)
            flush_text_input_events()
            # Match native `grim_was_key_pressed(ENTER)` after the input flush.
            rl.is_key_pressed(rl.KeyboardKey.KEY_ENTER)
            rl.is_key_pressed(rl.KeyboardKey.KEY_KP_ENTER)
            if idx < TABLE_MAX:
                self.phase = 0
                self.input_text = player_name_default[:NAME_MAX_EDIT]
                self.input_caret = len(self.input_text)
                self._defer_name_input_until_controls_released = True
                return None
            else:
                self.phase = 1

        # Basic text input behavior for the name-entry phase.
        if self.phase == 0:
            if self._defer_name_input_until_controls_released:
                flush_text_input_events()
                rl.is_key_pressed(rl.KeyboardKey.KEY_ENTER)
                rl.is_key_pressed(rl.KeyboardKey.KEY_KP_ENTER)
                if not gameplay_controls_held(self.config):
                    self._defer_name_input_until_controls_released = False
                return None
            click = rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
            typed = poll_text_input(NAME_MAX_EDIT - len(self.input_text), allow_space=True)
            if typed:
                self.input_text = (self.input_text[: self.input_caret] + typed + self.input_text[self.input_caret :])[
                    :NAME_MAX_EDIT
                ]
                self.input_caret = min(len(self.input_text), self.input_caret + len(typed))
                if play_sfx is not None:
                    play_sfx("sfx_ui_typeclick_01" if (int(rand()) & 1) == 0 else "sfx_ui_typeclick_02")
            if rl.is_key_pressed(rl.KeyboardKey.KEY_BACKSPACE):
                if self.input_caret > 0:
                    self.input_text = self.input_text[: self.input_caret - 1] + self.input_text[self.input_caret :]
                    self.input_caret -= 1
                    if play_sfx is not None:
                        play_sfx("sfx_ui_typeclick_01" if (int(rand()) & 1) == 0 else "sfx_ui_typeclick_02")
            if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT):
                self.input_caret = max(0, self.input_caret - 1)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT):
                self.input_caret = min(len(self.input_text), self.input_caret + 1)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_HOME):
                self.input_caret = 0
            if rl.is_key_pressed(rl.KeyboardKey.KEY_END):
                self.input_caret = len(self.input_text)

            screen_w = float(rl.get_screen_width())
            screen_h = float(rl.get_screen_height())
            scale = ui_scale(screen_w, screen_h)
            panel_layout = self._panel_layout(screen_w=screen_w, scale=scale)
            banner_pos = panel_layout.top_left + Vec2(GAME_OVER_BANNER_X_OFFSET * scale, 40.0 * scale)
            form_pos = banner_pos + Vec2(8.0 * scale, 84.0 * scale)
            ok_pos = form_pos + Vec2(170.0 * scale, 32.0 * scale)
            ok_w = button_width(self.font, self._ok_button.label, scale=scale, force_wide=self._ok_button.force_wide)
            ok_clicked = button_update(self._ok_button, pos=ok_pos, width=ok_w, dt_ms=dt_ms, mouse=mouse, click=click)

            if ok_clicked or rl.is_key_pressed(rl.KeyboardKey.KEY_ENTER):
                if self.input_text.strip():
                    if play_sfx is not None:
                        play_sfx("sfx_ui_typeenter")
                    candidate = (self._candidate_record or record).copy()
                    candidate.set_name(self.input_text)
                    self.config.set_player_name(self.input_text)
                    self.config.save()
                    path = scores_path_for_config(self.base_dir, self.config)
                    if not self._saved:
                        upsert_highscore_record(path, candidate)
                        self._saved = True
                    self.phase = 1
                    return None
                if play_sfx is not None:
                    play_sfx("sfx_shock_hit_01")
        else:
            # Buttons phase: let the caller handle navigation; we just report actions.
            click = rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
            screen_w = float(rl.get_screen_width())
            screen_h = float(rl.get_screen_height())
            scale = ui_scale(screen_w, screen_h)
            panel_layout = self._panel_layout(screen_w=screen_w, scale=scale)
            banner_pos = panel_layout.top_left + Vec2(GAME_OVER_BANNER_X_OFFSET * scale, 40.0 * scale)
            button_pos = banner_pos + Vec2(52.0 * scale, (210.0 if self.rank < TABLE_MAX else 208.0) * scale)

            play_again_w = button_width(
                self.font, self._play_again_button.label, scale=scale, force_wide=self._play_again_button.force_wide,
            )
            if button_update(
                self._play_again_button,
                pos=button_pos,
                width=play_again_w,
                dt_ms=dt_ms,
                mouse=mouse,
                click=click,
            ):
                if play_sfx is not None:
                    play_sfx("sfx_ui_buttonclick")
                self._begin_close_transition("play_again")
                return None
            button_pos = button_pos.offset(dy=32.0 * scale)

            high_scores_w = button_width(
                self.font, self._high_scores_button.label, scale=scale, force_wide=self._high_scores_button.force_wide,
            )
            if button_update(
                self._high_scores_button,
                pos=button_pos,
                width=high_scores_w,
                dt_ms=dt_ms,
                mouse=mouse,
                click=click,
            ):
                if play_sfx is not None:
                    play_sfx("sfx_ui_buttonclick")
                self._begin_close_transition("high_scores")
                return None
            button_pos = button_pos.offset(dy=32.0 * scale)

            main_menu_w = button_width(
                self.font, self._main_menu_button.label, scale=scale, force_wide=self._main_menu_button.force_wide,
            )
            if button_update(
                self._main_menu_button,
                pos=button_pos,
                width=main_menu_w,
                dt_ms=dt_ms,
                mouse=mouse,
                click=click,
            ):
                if play_sfx is not None:
                    play_sfx("sfx_ui_buttonclick")
                self._begin_close_transition("main_menu")
                return None
        return None

    def _draw_score_card(
        self,
        *,
        pos: Vec2,
        record: HighScoreRecord,
        hud_assets: HudAssets | None,
        alpha: float,
        show_weapon_row: bool,
        scale: float,
        mouse: rl.Vector2,
    ) -> None:
        dt_hover = float(self._dt) * 2.0
        label_color = rl.Color(COLOR_SCORE_LABEL.r, COLOR_SCORE_LABEL.g, COLOR_SCORE_LABEL.b, int(255 * alpha * 0.8))
        value_color = rl.Color(COLOR_SCORE_VALUE.r, COLOR_SCORE_VALUE.g, COLOR_SCORE_VALUE.b, int(255 * alpha))
        hint_color = rl.Color(COLOR_SCORE_LABEL.r, COLOR_SCORE_LABEL.g, COLOR_SCORE_LABEL.b, int(255 * alpha * 0.7))

        card_origin = pos.offset(dx=4.0 * scale)
        mode_raw = int(record.game_mode_id)
        try:
            mode_id = GameMode(mode_raw)
        except ValueError:
            mode_id = GameMode.DEMO

        # Left column: Score + value + Rank.
        score_label = "Score"
        score_label_w = self._text_width(score_label, 1.0 * scale)
        self._draw_small(
            score_label,
            card_origin.offset(dx=32.0 * scale - score_label_w * 0.5),
            1.0 * scale,
            label_color,
        )

        match mode_id:
            case GameMode.RUSH | GameMode.QUESTS:
                seconds = float(int(record.survival_elapsed_ms)) * 0.001
                score_value = f"{seconds:.2f} secs"
            case _:
                score_value = f"{int(record.score_xp)}"
        score_value_w = self._text_width(score_value, 1.0 * scale)
        self._draw_small(
            score_value,
            card_origin + Vec2(32.0 * scale - score_value_w * 0.5, 15.0 * scale),
            1.0 * scale,
            value_color,
        )

        rank_value = format_ordinal(int(self.rank) + 1)
        rank_text = f"Rank: {rank_value}"
        rank_w = self._text_width(rank_text, 1.0 * scale)
        self._draw_small(
            rank_text,
            card_origin + Vec2(32.0 * scale - rank_w * 0.5, 30.0 * scale),
            1.0 * scale,
            label_color,
        )

        # Separator between columns (mirrors FUN_00441220 + offset adjustments).
        separator_x = card_origin.x + 80.0 * scale
        rl.draw_line(
            int(separator_x),
            int(card_origin.y),
            int(separator_x),
            int(card_origin.y + 48.0 * scale),
            label_color,
        )

        # Right column: Game time + gauge, or Experience in quest mode.
        col2_pos = card_origin.offset(dx=96.0 * scale)
        match mode_id:
            case GameMode.QUESTS:
                self._draw_small("Experience", col2_pos, 1.0 * scale, label_color)
                xp_value = f"{int(record.score_xp)}"
                xp_w = self._text_width(xp_value, 1.0 * scale)
                self._draw_small(
                    xp_value,
                    col2_pos + Vec2(32.0 * scale - xp_w * 0.5, 15.0 * scale),
                    1.0 * scale,
                    label_color,
                )
                self._hover_time = max(0.0, float(self._hover_time) - dt_hover)
            case _:
                self._draw_small("Game time", col2_pos.offset(dx=6.0 * scale), 1.0 * scale, label_color)
                time_rect_pos = col2_pos + Vec2(8.0 * scale, 16.0 * scale)
                time_rect = Rect.from_top_left(time_rect_pos, 64.0 * scale, 29.0 * scale)
                hovering_time = time_rect.contains(mouse)
                self._hover_time = float(max(0.0, min(1.0, self._hover_time + (dt_hover if hovering_time else -dt_hover))))

                elapsed_ms = int(record.survival_elapsed_ms)
                if hud_assets is not None and hud_assets.clock_table is not None:
                    src = rl.Rectangle(0.0, 0.0, float(hud_assets.clock_table.width), float(hud_assets.clock_table.height))
                    clock_table_pos = col2_pos + Vec2(8.0 * scale, 14.0 * scale)
                    dst = rl.Rectangle(clock_table_pos.x, clock_table_pos.y, 32.0 * scale, 32.0 * scale)
                    rl.draw_texture_pro(
                        hud_assets.clock_table,
                        src,
                        dst,
                        rl.Vector2(0.0, 0.0),
                        0.0,
                        rl.Color(255, 255, 255, int(255 * alpha)),
                    )
                if hud_assets is not None and hud_assets.clock_pointer is not None:
                    src = rl.Rectangle(
                        0.0, 0.0, float(hud_assets.clock_pointer.width), float(hud_assets.clock_pointer.height),
                    )
                    # NOTE: Raylib's draw_texture_pro uses dst.x/y as the rotation origin position;
                    # offset by half-size so the 32x32 quad stays aligned with the table.
                    clock_pointer_pos = col2_pos + Vec2(24.0 * scale, 30.0 * scale)
                    dst = rl.Rectangle(clock_pointer_pos.x, clock_pointer_pos.y, 32.0 * scale, 32.0 * scale)
                    seconds = max(0, elapsed_ms // 1000)
                    rotation = float(seconds) * 6.0
                    origin = rl.Vector2(16.0 * scale, 16.0 * scale)
                    rl.draw_texture_pro(
                        hud_assets.clock_pointer, src, dst, origin, rotation, rl.Color(255, 255, 255, int(255 * alpha)),
                    )

                time_text = format_time_mm_ss(elapsed_ms)
                self._draw_small(time_text, col2_pos + Vec2(40.0 * scale, 19.0 * scale), 1.0 * scale, label_color)

        # Second row: weapon icon + frags + hit ratio (suppressed while entering the name).
        row_pos = card_origin.offset(dy=52.0 * scale)
        self._hover_weapon = float(max(0.0, min(1.0, self._hover_weapon)))
        self._hover_hit_ratio = float(max(0.0, min(1.0, self._hover_hit_ratio)))
        if show_weapon_row and hud_assets is not None and hud_assets.wicons is not None:
            weapon_pos = row_pos
            weapon_rect = Rect.from_top_left(weapon_pos, 64.0 * scale, 32.0 * scale)
            hovering_weapon = weapon_rect.contains(mouse)
            self._hover_weapon = float(
                max(0.0, min(1.0, self._hover_weapon + (dt_hover if hovering_weapon else -dt_hover))),
            )

            src = _weapon_icon_src(hud_assets.wicons, record.most_used_weapon_id)
            if src is not None:
                dst = rl.Rectangle(weapon_pos.x, weapon_pos.y, 64.0 * scale, 32.0 * scale)
                rl.draw_texture_pro(
                    hud_assets.wicons, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.Color(255, 255, 255, int(255 * alpha)),
                )

            weapon_id = record.most_used_weapon_id
            weapon_name = weapon_display_name(weapon_id, preserve_bugs=bool(self.preserve_bugs))
            name_w = self._text_width(weapon_name, 1.0 * scale)
            name_pos = Vec2(card_origin.x + max(0.0, (32.0 * scale - name_w * 0.5)), row_pos.y + 32.0 * scale)
            self._draw_small(weapon_name, name_pos, 1.0 * scale, hint_color)

            frags_text = f"Frags: {int(record.creature_kill_count)}"
            stats_pos = row_pos.offset(dx=110.0 * scale)
            self._draw_small(frags_text, stats_pos.offset(dy=1.0 * scale), 1.0 * scale, label_color)

            fired = max(0, int(record.shots_fired))
            hit = max(0, int(record.shots_hit))
            ratio = int((hit * 100) / fired) if fired > 0 else 0
            hit_text = f"Hit %: {ratio}%"
            self._draw_small(hit_text, stats_pos.offset(dy=15.0 * scale), 1.0 * scale, label_color)

            hit_rect_pos = stats_pos.offset(dy=15.0 * scale)
            hit_rect = Rect.from_top_left(hit_rect_pos, 64.0 * scale, 17.0 * scale)
            hovering_hit = hit_rect.contains(mouse)
            self._hover_hit_ratio = float(
                max(0.0, min(1.0, self._hover_hit_ratio + (dt_hover if hovering_hit else -dt_hover))),
            )
            tooltip_pos = row_pos.offset(dy=48.0 * scale)
        else:
            self._hover_weapon = max(0.0, float(self._hover_weapon) - dt_hover)
            self._hover_hit_ratio = 0.0
            tooltip_pos = row_pos

        self._hover_weapon = float(max(0.0, min(1.0, self._hover_weapon)))
        self._hover_time = float(max(0.0, min(1.0, self._hover_time)))
        self._hover_hit_ratio = float(max(0.0, min(1.0, self._hover_hit_ratio)))

        if self._hover_weapon > 0.5:
            t = (self._hover_weapon - 0.5) * 2.0
            col = rl.Color(label_color.r, label_color.g, label_color.b, int(255 * alpha * t))
            self._draw_small(
                "Most used weapon during the game",
                tooltip_pos.offset(dx=-20.0 * scale),
                1.0 * scale,
                col,
            )
        if self._hover_time > 0.5:
            t = (self._hover_time - 0.5) * 2.0
            col = rl.Color(label_color.r, label_color.g, label_color.b, int(255 * alpha * t))
            self._draw_small(
                "The time the game lasted",
                tooltip_pos.offset(dx=12.0 * scale),
                1.0 * scale,
                col,
            )
        if self._hover_hit_ratio > 0.5:
            t = (self._hover_hit_ratio - 0.5) * 2.0
            col = rl.Color(label_color.r, label_color.g, label_color.b, int(255 * alpha * t))
            hit_ratio_tooltip = (
                "The % of shot bullets hit the target"
                if bool(self.preserve_bugs)
                else "The % of bullets that hit the target"
            )
            self._draw_small(
                hit_ratio_tooltip,
                tooltip_pos.offset(dx=-22.0 * scale),
                1.0 * scale,
                col,
            )

    def draw(
        self,
        *,
        record: HighScoreRecord,
        banner_kind: str,
        hud_assets: HudAssets | None,
        mouse: rl.Vector2 | None = None,
    ) -> None:
        if self.assets is None:
            return
        if mouse is None:
            mouse = rl.get_mouse_position()

        screen_w = float(rl.get_screen_width())
        screen_h = float(rl.get_screen_height())
        scale = ui_scale(screen_w, screen_h)

        panel_layout = self._panel_layout(screen_w=screen_w, scale=scale)
        panel = panel_layout.panel
        panel_top_left = panel_layout.top_left

        # Panel background
        if self.assets.menu_panel is not None:
            fx_detail = self.config.fx_detail(level=0, default=False)
            draw_classic_menu_panel(
                self.assets.menu_panel,
                dst=panel.to_rl(),
                tint=rl.WHITE,
                shadow=fx_detail,
            )

        # Banner (Reaper / Well done)
        banner_pos = panel_top_left + Vec2(GAME_OVER_BANNER_X_OFFSET * scale, 40.0 * scale)
        banner = self.assets.text_reaper if banner_kind == "reaper" else self.assets.text_well_done
        if banner is not None:
            _draw_texture_centered(
                banner,
                banner_pos,
                TEXTURE_TOP_BANNER_W * scale,
                TEXTURE_TOP_BANNER_H * scale,
                1.0,
            )

        if self.phase == 0:
            form_pos = banner_pos + Vec2(8.0 * scale, 84.0 * scale)
            self._draw_small("State your name, trooper!", form_pos.offset(dx=42.0 * scale), 1.0 * scale, COLOR_TEXT)

            input_pos = form_pos.offset(dy=40.0 * scale)
            rl.draw_rectangle_lines(
                int(input_pos.x), int(input_pos.y), int(INPUT_BOX_W * scale), int(INPUT_BOX_H * scale), rl.WHITE,
            )
            rl.draw_rectangle(
                int(input_pos.x + 1.0 * scale),
                int(input_pos.y + 1.0 * scale),
                int((INPUT_BOX_W - 2.0) * scale),
                int((INPUT_BOX_H - 2.0) * scale),
                rl.Color(0, 0, 0, 255),
            )
            draw_ui_text(
                self.font,
                self.input_text,
                input_pos + Vec2(4.0 * scale, 2.0 * scale),
                scale=1.0 * scale,
                color=COLOR_TEXT_MUTED,
            )
            caret_alpha = 1.0
            if math.sin(float(rl.get_time()) * 4.0) > 0.0:
                caret_alpha = 0.4
            caret_color = rl.Color(255, 255, 255, int(255 * caret_alpha))
            caret_x = input_pos.x + 4.0 * scale + self._text_width(self.input_text[: self.input_caret], 1.0 * scale)
            rl.draw_rectangle(
                int(caret_x), int(input_pos.y + 2.0 * scale), int(1.0 * scale), int(14.0 * scale), caret_color,
            )

            ok_pos = form_pos + Vec2(170.0 * scale, 32.0 * scale)
            ok_w = button_width(self.font, self._ok_button.label, scale=scale, force_wide=self._ok_button.force_wide)
            button_draw(
                self.assets.perk_menu_assets,
                self.font,
                self._ok_button,
                pos=ok_pos,
                width=ok_w,
                scale=scale,
            )

            score_pos = form_pos + Vec2(16.0 * scale, 116.0 * scale)
            self._draw_score_card(
                pos=score_pos,
                record=record,
                hud_assets=hud_assets,
                alpha=1.0,
                show_weapon_row=False,
                scale=scale,
                mouse=mouse,
            )
        else:
            score_card_pos = banner_pos + Vec2(
                30.0 * scale,
                (80.0 if self.rank < TABLE_MAX else 78.0) * scale,
            )
            if self.rank >= TABLE_MAX and banner_kind == "reaper":
                self._draw_small(
                    "Score too low for top100.",
                    banner_pos + Vec2(38.0 * scale, 62.0 * scale),
                    1.0 * scale,
                    rl.Color(200, 200, 200, 255),
                )

            self._draw_score_card(
                pos=score_card_pos,
                record=record,
                hud_assets=hud_assets,
                alpha=1.0,
                show_weapon_row=True,
                scale=scale,
                mouse=mouse,
            )

        # Buttons phase rendering.
        if self.phase == 1:
            button_pos = banner_pos + Vec2(52.0 * scale, (210.0 if self.rank < TABLE_MAX else 208.0) * scale)
            play_again_w = button_width(
                self.font, self._play_again_button.label, scale=scale, force_wide=self._play_again_button.force_wide,
            )
            button_draw(
                self.assets.perk_menu_assets,
                self.font,
                self._play_again_button,
                pos=button_pos,
                width=play_again_w,
                scale=scale,
            )
            button_pos = button_pos.offset(dy=32.0 * scale)

            high_scores_w = button_width(
                self.font, self._high_scores_button.label, scale=scale, force_wide=self._high_scores_button.force_wide,
            )
            button_draw(
                self.assets.perk_menu_assets,
                self.font,
                self._high_scores_button,
                pos=button_pos,
                width=high_scores_w,
                scale=scale,
            )
            button_pos = button_pos.offset(dy=32.0 * scale)

            main_menu_w = button_width(
                self.font, self._main_menu_button.label, scale=scale, force_wide=self._main_menu_button.force_wide,
            )
            button_draw(
                self.assets.perk_menu_assets,
                self.font,
                self._main_menu_button,
                pos=button_pos,
                width=main_menu_w,
                scale=scale,
            )

        draw_menu_cursor(
            self.assets.particles,
            self.assets.perk_menu_assets.cursor,
            pos=Vec2.from_xy(mouse),
            pulse_time=float(self._cursor_pulse_time),
        )



from __future__ import annotations

from tqdm import tqdm

from . import dbg as _dbg
from . import net as _net
from . import relay as _relay
from . import replay as _replay
from . import root as _root

app = _root.app
replay_app = _replay.replay_app
dbg_app = _dbg.dbg_app
net_app = _net.net_app
relay_app = _relay.relay_app

app.add_typer(replay_app, name="replay")
app.add_typer(dbg_app, name="dbg")
app.add_typer(net_app, name="net")
app.add_typer(relay_app, name="relay")

def _replay_render_progress_callback(*, total_ticks: int, render_audio: bool):
    return _replay._replay_render_progress_callback(
        total_ticks=total_ticks,
        render_audio=render_audio,
        tqdm_factory=tqdm,
    )

def main(argv: list[str] | None = None) -> None:
    app(prog_name="crimson", args=argv)



from __future__ import annotations

from enum import Enum


class RtxRenderMode(str, Enum):
    CLASSIC = "classic"
    RTX = "rtx"


def parse_rtx_render_mode(raw: str) -> RtxRenderMode:
    value = str(raw).strip().lower()
    if value == "classic":
        return RtxRenderMode.CLASSIC
    if value == "rtx":
        return RtxRenderMode.RTX
    raise ValueError(f"unsupported render mode {raw!r}; expected classic|rtx")


def mode_from_rtx_flag(enabled: bool) -> RtxRenderMode:
    if enabled:
        return RtxRenderMode.RTX
    return RtxRenderMode.CLASSIC


def cycle_rtx_render_mode(mode: RtxRenderMode) -> RtxRenderMode:
    if mode is RtxRenderMode.RTX:
        return RtxRenderMode.CLASSIC
    return RtxRenderMode.RTX


__all__ = [
    "RtxRenderMode",
    "cycle_rtx_render_mode",
    "mode_from_rtx_flag",
    "parse_rtx_render_mode",
]



from __future__ import annotations

import msgspec

from ..effects import ParticleStyleId
from ..projectiles.types import ProjectileTemplateId, SecondaryProjectileTypeId
from ..weapons import WeaponId, projectile_type_id_for_weapon_id


class NoJitter(msgspec.Struct, frozen=True, tag=True):
    pass


class ModuloCenteredJitter(msgspec.Struct, frozen=True, tag=True):
    modulo: int
    center: int
    step: float


class MaskCenteredJitter(msgspec.Struct, frozen=True, tag=True):
    mask: int
    center: int
    step: float


type PelletJitterRule = NoJitter | ModuloCenteredJitter | MaskCenteredJitter


class NoSpeedScale(msgspec.Struct, frozen=True, tag=True):
    pass


class ModuloSpeedScale(msgspec.Struct, frozen=True, tag=True):
    base: float
    modulo: int
    step: float


type SpeedScaleRule = NoSpeedScale | ModuloSpeedScale


class NoTargetHint(msgspec.Struct, frozen=True, tag=True):
    pass


class UseAimTargetHint(msgspec.Struct, frozen=True, tag=True):
    pass


type SecondaryTargetingPolicy = NoTargetHint | UseAimTargetHint


class PrimaryPelletsMode(msgspec.Struct, frozen=True, tag=True):
    type_id: ProjectileTemplateId | None = None
    count: int | None = None
    jitter: PelletJitterRule = msgspec.field(default_factory=NoJitter)
    speed_scale: SpeedScaleRule = msgspec.field(default_factory=NoSpeedScale)


class SecondaryShotMode(msgspec.Struct, frozen=True, tag=True):
    type_id: SecondaryProjectileTypeId
    targeting: SecondaryTargetingPolicy = msgspec.field(default_factory=NoTargetHint)


class ParticleStreamMode(msgspec.Struct, frozen=True, tag=True):
    style: ParticleStyleId | None = None
    slow: bool = False


class MultiPlasmaFanMode(msgspec.Struct, frozen=True, tag=True):
    pass


class SwarmerDumpMode(msgspec.Struct, frozen=True, tag=True):
    pass


type FireMode = PrimaryPelletsMode | SecondaryShotMode | ParticleStreamMode | MultiPlasmaFanMode | SwarmerDumpMode


class FireRecipe(msgspec.Struct, frozen=True):
    mode: FireMode
    ammo_cost: float = 1.0


def pellet_jitter_step_for_weapon(weapon_id: WeaponId) -> float:
    weapon_id = WeaponId(weapon_id)
    if weapon_id == WeaponId.SHOTGUN:
        return 0.0013
    if weapon_id == WeaponId.SAWED_OFF_SHOTGUN:
        return 0.004
    if weapon_id == WeaponId.JACKHAMMER:
        return 0.0013
    return 0.0015


_SHOTGUN_SPEED_RANDOMIZE_WEAPONS: frozenset[WeaponId] = frozenset(
    {
        WeaponId.SHOTGUN,
        WeaponId.SAWED_OFF_SHOTGUN,
        WeaponId.JACKHAMMER,
    },
)


_DEFAULT_SPREAD_JITTER = ModuloCenteredJitter(modulo=200, center=100, step=0.0015)
_DEFAULT_SPEED_SCALE = ModuloSpeedScale(base=1.0, modulo=100, step=0.01)
_GAUSS_ION_SPEED_SCALE = ModuloSpeedScale(base=1.4, modulo=0x50, step=0.01)


FIRE_RECIPE_BY_WEAPON: dict[WeaponId, FireRecipe] = {
    WeaponId.ROCKET_LAUNCHER: FireRecipe(mode=SecondaryShotMode(type_id=SecondaryProjectileTypeId.ROCKET)),
    WeaponId.SEEKER_ROCKETS: FireRecipe(
        mode=SecondaryShotMode(
            type_id=SecondaryProjectileTypeId.HOMING_ROCKET,
            targeting=UseAimTargetHint(),
        ),
    ),
    WeaponId.ROCKET_MINIGUN: FireRecipe(mode=SecondaryShotMode(type_id=SecondaryProjectileTypeId.ROCKET_MINIGUN)),
    WeaponId.FLAMETHROWER: FireRecipe(
        mode=ParticleStreamMode(style=None, slow=False),
        ammo_cost=0.1,
    ),
    WeaponId.BLOW_TORCH: FireRecipe(
        mode=ParticleStreamMode(style=ParticleStyleId.BLOW_TORCH, slow=False),
        ammo_cost=0.05,
    ),
    WeaponId.HR_FLAMER: FireRecipe(
        mode=ParticleStreamMode(style=ParticleStyleId.HR_FLAMER, slow=False),
        ammo_cost=0.1,
    ),
    WeaponId.BUBBLEGUN: FireRecipe(
        mode=ParticleStreamMode(style=None, slow=True),
        ammo_cost=0.15,
    ),
    WeaponId.MULTI_PLASMA: FireRecipe(mode=MultiPlasmaFanMode()),
    WeaponId.MINI_ROCKET_SWARMERS: FireRecipe(mode=SwarmerDumpMode()),
    WeaponId.PLASMA_SHOTGUN: FireRecipe(
        mode=PrimaryPelletsMode(
            type_id=ProjectileTemplateId.PLASMA_MINIGUN,
            count=14,
            jitter=MaskCenteredJitter(mask=0xFF, center=0x80, step=0.002),
            speed_scale=_DEFAULT_SPEED_SCALE,
        ),
    ),
    WeaponId.GAUSS_SHOTGUN: FireRecipe(
        mode=PrimaryPelletsMode(
            type_id=ProjectileTemplateId.GAUSS_GUN,
            count=6,
            jitter=ModuloCenteredJitter(modulo=200, center=100, step=0.002),
            speed_scale=_GAUSS_ION_SPEED_SCALE,
        ),
    ),
    WeaponId.ION_SHOTGUN: FireRecipe(
        mode=PrimaryPelletsMode(
            type_id=ProjectileTemplateId.ION_MINIGUN,
            count=8,
            jitter=ModuloCenteredJitter(modulo=200, center=100, step=0.0026),
            speed_scale=_GAUSS_ION_SPEED_SCALE,
        ),
    ),
}


def resolve_fire_recipe(
    weapon_id: WeaponId,
    *,
    pellet_count: int,
    fire_bullets_active: bool,
) -> FireRecipe:
    if fire_bullets_active:
        return FireRecipe(
            mode=PrimaryPelletsMode(
                type_id=ProjectileTemplateId.FIRE_BULLETS,
                count=max(0, int(pellet_count)),
                jitter=_DEFAULT_SPREAD_JITTER,
                speed_scale=NoSpeedScale(),
            ),
        )

    recipe = FIRE_RECIPE_BY_WEAPON.get(WeaponId(weapon_id))
    if recipe is not None:
        return recipe

    pellets = max(1, int(pellet_count))
    jitter: PelletJitterRule = NoJitter()
    if pellets > 1:
        jitter = ModuloCenteredJitter(
            modulo=200,
            center=100,
            step=pellet_jitter_step_for_weapon(weapon_id),
        )

    speed_scale: SpeedScaleRule = NoSpeedScale()
    if pellets > 1 and WeaponId(weapon_id) in _SHOTGUN_SPEED_RANDOMIZE_WEAPONS:
        speed_scale = _DEFAULT_SPEED_SCALE

    return FireRecipe(
        mode=PrimaryPelletsMode(
            type_id=projectile_type_id_for_weapon_id(WeaponId(weapon_id)),
            count=pellets,
            jitter=jitter,
            speed_scale=speed_scale,
        ),
    )


__all__ = [
    "FIRE_RECIPE_BY_WEAPON",
    "FireMode",
    "FireRecipe",
    "MaskCenteredJitter",
    "ModuloCenteredJitter",
    "ModuloSpeedScale",
    "MultiPlasmaFanMode",
    "NoJitter",
    "NoSpeedScale",
    "NoTargetHint",
    "ParticleStreamMode",
    "PelletJitterRule",
    "PrimaryPelletsMode",
    "SecondaryShotMode",
    "SecondaryTargetingPolicy",
    "SpeedScaleRule",
    "SwarmerDumpMode",
    "UseAimTargetHint",
    "pellet_jitter_step_for_weapon",
    "resolve_fire_recipe",
]



from __future__ import annotations

import math

from ...owner_ref import OwnerRef
from ...projectiles.types import ProjectileTemplateId
from ..helpers import perk_active
from ..ids import PerkId
from ..runtime.hook_types import PerkHooks
from ..runtime.player_tick_context import PlayerPerkTickCtx


def tick_hot_tempered(ctx: PlayerPerkTickCtx) -> None:
    if not perk_active(ctx.player, PerkId.HOT_TEMPERED):
        ctx.player.hot_tempered_timer = 0.0
        return

    ctx.player.hot_tempered_timer += ctx.dt
    if ctx.player.hot_tempered_timer <= ctx.state.perk_intervals.hot_tempered:
        return

    owner = (
        ctx.owner_ref_for_player(ctx.player.index)
        if ctx.state.friendly_fire_enabled
        else OwnerRef.from_local_player(0)
    )
    for idx in range(8):
        type_id = ProjectileTemplateId.PLASMA_MINIGUN if ((idx & 1) == 0) else ProjectileTemplateId.PLASMA_RIFLE
        angle = float(idx) * (math.pi / 4.0)
        ctx.projectile_spawn(
            ctx.state,
            players=ctx.players,
            pos=ctx.player_pos_before_move,
            angle=angle,
            type_id=type_id,
            owner=owner,
            owner_player_index=ctx.player.index,
        )
    ctx.state.sfx_queue.append("sfx_explosion_small")

    ctx.player.hot_tempered_timer -= ctx.state.perk_intervals.hot_tempered
    ctx.state.perk_intervals.hot_tempered = float(ctx.state.rng.rand() % 8) + 2.0


HOOKS = PerkHooks(
    perk_id=PerkId.HOT_TEMPERED,
    player_tick_steps=(tick_hot_tempered,),
)



from __future__ import annotations

from ..game_modes import GameMode
from ..net.lockstep_protocol import INPUT_DELAY_TICKS as LOCKSTEP_INPUT_DELAY_TICKS
from ..net.session_settings import LockstepSessionSettings, session_settings_for_lockstep
from .types import BootstrapKind, ReplayHeader, ReplayStatusSnapshot


def session_settings_from_replay_header(
    header: ReplayHeader,
    *,
    input_delay_ticks: int = LOCKSTEP_INPUT_DELAY_TICKS,
) -> LockstepSessionSettings:
    return session_settings_for_lockstep(
        mode_id=int(header.game_mode_id),
        player_count=int(header.player_count),
        quest_level=str(header.quest_level),
        preserve_bugs=bool(header.preserve_bugs),
        tick_rate=int(header.tick_rate),
        input_delay_ticks=int(input_delay_ticks),
    )


def replay_header_from_session_settings(
    settings: LockstepSessionSettings,
    *,
    seed: int,
    bootstrap_kind: BootstrapKind = "none",
    bootstrap_seed: int = 0,
    quest_fail_retry_count: int = 0,
    hardcore: bool = False,
    detail_preset: int = 5,
    gore_disabled: int = 0,
    world_size: float = 1024.0,
    status: ReplayStatusSnapshot | None = None,
) -> ReplayHeader:
    mode_raw = int(settings.mode_id)
    try:
        game_mode_id = GameMode(mode_raw)
    except ValueError as exc:
        raise ValueError(f"unsupported replay game_mode_id={mode_raw}") from exc

    return ReplayHeader(
        game_mode_id=game_mode_id,
        seed=int(seed),
        quest_level=str(settings.quest_level),
        bootstrap_kind=bootstrap_kind,
        bootstrap_seed=int(bootstrap_seed),
        tick_rate=int(settings.tick_rate),
        quest_fail_retry_count=int(quest_fail_retry_count),
        hardcore=bool(hardcore),
        preserve_bugs=bool(settings.preserve_bugs),
        detail_preset=int(detail_preset),
        gore_disabled=int(gore_disabled),
        world_size=float(world_size),
        player_count=int(settings.player_count),
        status=ReplayStatusSnapshot() if status is None else status,
    )



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path

from grim.render_pipeline import RenderPresent, RenderSink, WindowSink

__all__ = ["RenderPresent", "RenderSink", "WindowSink", "NullSink", "VideoSink"]


class NullSink:
    """Headless sink used for determinism-only verification paths."""

    def open(self) -> None:
        pass

    def present(self) -> None:
        pass

    def flush(self) -> None:
        pass

    def close(self) -> None:
        pass


class VideoSink:
    """Video-export sink. Fail-fast by policy: presentation errors abort rendering."""

    def __init__(
        self,
        *,
        output_path: Path,
        open_transport: Callable[[], None] | None = None,
        present_frame: RenderPresent | None = None,
        flush_transport: Callable[[], None] | None = None,
        close_transport: Callable[[], None] | None = None,
    ) -> None:
        self.output_path = Path(output_path)
        self._open_transport = open_transport
        self._present_frame = present_frame
        self._flush_transport = flush_transport
        self._close_transport = close_transport
        self._opened = False
        self._flushed = False

    def open(self) -> None:
        self.output_path.parent.mkdir(parents=True, exist_ok=True)
        if self._open_transport is not None:
            self._open_transport()
        self._opened = True
        self._flushed = False

    def present(self) -> None:
        if self._present_frame is not None:
            self._present_frame()

    def flush(self) -> None:
        if not self._opened or self._flushed:
            return
        if self._flush_transport is not None:
            self._flush_transport()
        self._flushed = True

    def close(self) -> None:
        if not self._opened:
            return
        if self._close_transport is not None:
            self._close_transport()
        self._opened = False



from __future__ import annotations

from typing import Literal, TypeAlias

import msgspec

from .lockstep_protocol import INPUT_DELAY_TICKS as LOCKSTEP_INPUT_DELAY_TICKS
from .lockstep_protocol import PROTOCOL_VERSION as LOCKSTEP_PROTOCOL_VERSION
from .lockstep_protocol import TICK_RATE as LOCKSTEP_TICK_RATE
from .lockstep_protocol import Hello, MatchStart, Welcome
from .lockstep_protocol import StatusSnapshot as LockstepStatusSnapshot
from .relay_protocol import (
    INPUT_DELAY_TICKS as RELAY_INPUT_DELAY_TICKS,
)
from .relay_protocol import (
    ROLLBACK_MAX_TICKS,
    NetcodeMode,
    RelaySlot,
    RoomCreate,
    RoomStart,
    RoomState,
    StatusSnapshot,
)
from .relay_protocol import (
    TICK_RATE as RELAY_TICK_RATE,
)


class LockstepSessionSettings(msgspec.Struct, frozen=True):
    mode_id: int = 0
    player_count: int = 1
    quest_level: str = ""
    preserve_bugs: bool = False
    tick_rate: int = LOCKSTEP_TICK_RATE
    input_delay_ticks: int = LOCKSTEP_INPUT_DELAY_TICKS
    netcode_mode: Literal["lockstep"] = "lockstep"


class RelaySessionSettings(msgspec.Struct, frozen=True):
    mode_id: int = 0
    player_count: int = 1
    quest_level: str = ""
    preserve_bugs: bool = False
    tick_rate: int = RELAY_TICK_RATE
    input_delay_ticks: int = RELAY_INPUT_DELAY_TICKS
    rollback_max_ticks: int = ROLLBACK_MAX_TICKS
    netcode_mode: NetcodeMode = "rollback"


SessionSettings: TypeAlias = LockstepSessionSettings | RelaySessionSettings


def session_settings_from_hello(message: Hello) -> LockstepSessionSettings:
    return session_settings_for_lockstep(
        mode_id=int(message.mode_id),
        player_count=int(message.player_count),
        quest_level=str(message.quest_level),
        preserve_bugs=bool(message.preserve_bugs),
        tick_rate=int(message.tick_rate),
        input_delay_ticks=int(message.input_delay_ticks),
    )


def session_settings_for_lockstep(
    *,
    mode_id: int,
    player_count: int,
    quest_level: str,
    preserve_bugs: bool,
    tick_rate: int = LOCKSTEP_TICK_RATE,
    input_delay_ticks: int = LOCKSTEP_INPUT_DELAY_TICKS,
) -> LockstepSessionSettings:
    return LockstepSessionSettings(
        mode_id=int(mode_id),
        player_count=max(1, min(4, int(player_count))),
        quest_level=str(quest_level or ""),
        preserve_bugs=bool(preserve_bugs),
        tick_rate=max(1, int(tick_rate)),
        input_delay_ticks=max(0, int(input_delay_ticks)),
    )


def session_settings_from_welcome(message: Welcome) -> LockstepSessionSettings:
    return session_settings_for_lockstep(
        mode_id=int(message.mode_id),
        player_count=int(message.player_count),
        quest_level=str(message.quest_level),
        preserve_bugs=bool(message.preserve_bugs),
        tick_rate=int(message.tick_rate),
        input_delay_ticks=int(message.input_delay_ticks),
    )


def welcome_from_session_settings(
    settings: LockstepSessionSettings,
    *,
    accepted: bool,
    reason: str = "",
    session_id: str = "",
    protocol_version: int = LOCKSTEP_PROTOCOL_VERSION,
    build_id: str = "",
    slot_index: int = -1,
    host_slot_index: int = 0,
    seed: int = 0,
    started: bool = False,
) -> Welcome:
    return Welcome(
        accepted=bool(accepted),
        reason=str(reason),
        session_id=str(session_id),
        protocol_version=int(protocol_version),
        build_id=str(build_id),
        mode_id=int(settings.mode_id),
        player_count=int(settings.player_count),
        slot_index=int(slot_index),
        host_slot_index=int(host_slot_index),
        tick_rate=int(settings.tick_rate),
        input_delay_ticks=int(settings.input_delay_ticks),
        seed=int(seed),
        quest_level=str(settings.quest_level),
        preserve_bugs=bool(settings.preserve_bugs),
        started=bool(started),
    )


def session_settings_from_match_start(
    message: MatchStart,
    *,
    tick_rate: int = LOCKSTEP_TICK_RATE,
    input_delay_ticks: int = LOCKSTEP_INPUT_DELAY_TICKS,
) -> LockstepSessionSettings:
    return session_settings_for_lockstep(
        mode_id=int(message.mode_id),
        player_count=int(message.player_count),
        quest_level=str(message.quest_level),
        preserve_bugs=bool(message.preserve_bugs),
        tick_rate=int(tick_rate),
        input_delay_ticks=int(input_delay_ticks),
    )


def match_start_from_session_settings(
    settings: LockstepSessionSettings,
    *,
    session_id: str,
    seed: int,
    start_tick: int = 0,
    status_snapshot: LockstepStatusSnapshot | None = None,
) -> MatchStart:
    return MatchStart(
        session_id=str(session_id),
        mode_id=int(settings.mode_id),
        player_count=int(settings.player_count),
        seed=int(seed),
        start_tick=int(start_tick),
        quest_level=str(settings.quest_level),
        preserve_bugs=bool(settings.preserve_bugs),
        status_snapshot=status_snapshot,
    )


def session_settings_for_relay(
    *,
    mode_id: int,
    player_count: int,
    quest_level: str,
    preserve_bugs: bool,
    tick_rate: int = RELAY_TICK_RATE,
    input_delay_ticks: int = RELAY_INPUT_DELAY_TICKS,
    rollback_max_ticks: int = ROLLBACK_MAX_TICKS,
    netcode_mode: NetcodeMode = "rollback",
) -> RelaySessionSettings:
    return RelaySessionSettings(
        mode_id=int(mode_id),
        player_count=max(1, min(4, int(player_count))),
        quest_level=str(quest_level or ""),
        preserve_bugs=bool(preserve_bugs),
        tick_rate=max(1, int(tick_rate)),
        input_delay_ticks=max(0, int(input_delay_ticks)),
        rollback_max_ticks=max(1, int(rollback_max_ticks)),
        netcode_mode=netcode_mode,
    )


def session_settings_from_room_create(message: RoomCreate) -> RelaySessionSettings:
    return session_settings_for_relay(
        mode_id=int(message.mode_id),
        player_count=int(message.player_count),
        quest_level=str(message.quest_level),
        preserve_bugs=bool(message.preserve_bugs),
        tick_rate=int(message.tick_rate),
        input_delay_ticks=int(message.input_delay_ticks),
        rollback_max_ticks=int(message.rollback_max_ticks),
        netcode_mode=message.netcode_mode,
    )


def hello_from_session_settings(
    settings: LockstepSessionSettings,
    *,
    protocol_version: int,
    build_id: str,
    host: bool,
) -> Hello:
    return Hello(
        protocol_version=int(protocol_version),
        build_id=str(build_id),
        mode_id=int(settings.mode_id),
        player_count=int(settings.player_count),
        tick_rate=int(settings.tick_rate),
        input_delay_ticks=int(settings.input_delay_ticks),
        quest_level=str(settings.quest_level),
        preserve_bugs=bool(settings.preserve_bugs),
        host=bool(host),
    )


def room_create_from_session_settings(
    settings: RelaySessionSettings,
    *,
    status_snapshot: StatusSnapshot | None = None,
) -> RoomCreate:
    return RoomCreate(
        mode_id=int(settings.mode_id),
        player_count=int(settings.player_count),
        quest_level=str(settings.quest_level),
        preserve_bugs=bool(settings.preserve_bugs),
        tick_rate=int(settings.tick_rate),
        input_delay_ticks=int(settings.input_delay_ticks),
        rollback_max_ticks=int(settings.rollback_max_ticks),
        netcode_mode=settings.netcode_mode,
        status_snapshot=status_snapshot,
    )


def room_state_from_session_settings(
    settings: RelaySessionSettings,
    *,
    room_code: str,
    session_id: str,
    slots: list[RelaySlot],
    all_ready: bool,
    started: bool,
) -> RoomState:
    return RoomState(
        room_code=str(room_code),
        session_id=str(session_id),
        mode_id=int(settings.mode_id),
        player_count=int(settings.player_count),
        quest_level=str(settings.quest_level),
        preserve_bugs=bool(settings.preserve_bugs),
        tick_rate=int(settings.tick_rate),
        input_delay_ticks=int(settings.input_delay_ticks),
        rollback_max_ticks=int(settings.rollback_max_ticks),
        netcode_mode=settings.netcode_mode,
        slots=slots,
        all_ready=bool(all_ready),
        started=bool(started),
    )


def room_start_from_session_settings(
    settings: RelaySessionSettings,
    *,
    room_code: str,
    session_id: str,
    seed: int,
    start_tick: int,
    slot_index: int,
    host_slot_index: int,
    reconnect_token: str,
    status_snapshot: StatusSnapshot | None = None,
) -> RoomStart:
    return RoomStart(
        room_code=str(room_code),
        session_id=str(session_id),
        seed=int(seed),
        start_tick=int(start_tick),
        mode_id=int(settings.mode_id),
        player_count=int(settings.player_count),
        quest_level=str(settings.quest_level),
        preserve_bugs=bool(settings.preserve_bugs),
        tick_rate=int(settings.tick_rate),
        input_delay_ticks=int(settings.input_delay_ticks),
        rollback_max_ticks=int(settings.rollback_max_ticks),
        netcode_mode=settings.netcode_mode,
        slot_index=int(slot_index),
        host_slot_index=int(host_slot_index),
        reconnect_token=str(reconnect_token),
        status_snapshot=status_snapshot,
    )



from __future__ import annotations

from dataclasses import dataclass

QUEST_STAGE_COUNT = 5
QUESTS_PER_STAGE = 10
QUEST_COUNT = QUEST_STAGE_COUNT * QUESTS_PER_STAGE

# `game.cfg` quest counters use staged offsets into `quest_play_counts`.
QUEST_STATUS_GAMES_OFFSET = 11
QUEST_STATUS_COMPLETED_OFFSET = 51
QUEST_STATUS_TRACKED_COUNT = 40


@dataclass(frozen=True, order=True, slots=True)
class QuestLevel:
    major: int
    minor: int

    def __post_init__(self) -> None:
        major = int(self.major)
        minor = int(self.minor)
        object.__setattr__(self, "major", major)
        object.__setattr__(self, "minor", minor)
        if not (1 <= major <= QUEST_STAGE_COUNT):
            raise ValueError(f"quest stage out of range: {major} (expected 1..{QUEST_STAGE_COUNT})")
        if not (1 <= minor <= QUESTS_PER_STAGE):
            raise ValueError(f"quest row out of range: {minor} (expected 1..{QUESTS_PER_STAGE})")

    @classmethod
    def from_parts(cls, major: int, minor: int) -> QuestLevel:
        return cls(int(major), int(minor))

    @classmethod
    def from_parts_or_none(cls, major: int, minor: int) -> QuestLevel | None:
        major_i = int(major)
        minor_i = int(minor)
        if major_i == 0 and minor_i == 0:
            return None
        try:
            return cls(major_i, minor_i)
        except ValueError:
            return None

    @classmethod
    def parse(cls, value: str) -> QuestLevel:
        text = str(value).strip()
        major_text, sep, minor_text = text.partition(".")
        if sep != "." or not major_text or not minor_text:
            raise ValueError(f"invalid quest level: {value!r}")
        try:
            major = int(major_text)
            minor = int(minor_text)
        except ValueError as exc:
            raise ValueError(f"invalid quest level: {value!r}") from exc
        return cls(major=major, minor=minor)

    @classmethod
    def try_parse(cls, value: str | None) -> QuestLevel | None:
        text = str(value or "").strip()
        if not text:
            return None
        try:
            return cls.parse(text)
        except ValueError:
            return None

    @classmethod
    def from_global_index(cls, index: int) -> QuestLevel:
        idx = int(index)
        if not (0 <= idx < QUEST_COUNT):
            raise ValueError(f"quest global index out of range: {idx} (expected 0..{QUEST_COUNT - 1})")
        major, row_index = divmod(idx, QUESTS_PER_STAGE)
        return cls(major=major + 1, minor=row_index + 1)

    @classmethod
    def from_stage_row(cls, stage: int, row_index: int) -> QuestLevel:
        return cls(major=int(stage), minor=int(row_index) + 1)

    def to_string(self) -> str:
        return f"{self.major}.{self.minor}"

    @property
    def level(self) -> str:
        return self.to_string()

    def to_stage_pair(self) -> tuple[int, int]:
        return self.major, self.minor

    @property
    def stage_pair(self) -> tuple[int, int]:
        return self.to_stage_pair()

    def to_stage_row(self) -> tuple[int, int]:
        return self.major, self.minor - 1

    @property
    def stage_row(self) -> tuple[int, int]:
        return self.to_stage_row()

    def to_global_index(self) -> int:
        return (self.major - 1) * QUESTS_PER_STAGE + (self.minor - 1)

    @property
    def global_index(self) -> int:
        return self.to_global_index()

    def status_games_counter_index(self) -> int:
        return self.to_global_index() + QUEST_STATUS_GAMES_OFFSET

    @property
    def games_counter_index(self) -> int:
        return self.status_games_counter_index()

    def status_completed_counter_index(self) -> int:
        return self.to_global_index() + QUEST_STATUS_COMPLETED_OFFSET

    @property
    def completed_counter_index(self) -> int:
        return self.status_completed_counter_index()

    @property
    def tracked_in_status(self) -> bool:
        return self.to_global_index() < QUEST_STATUS_TRACKED_COUNT

    def tracked_status_games_counter_index(self) -> int | None:
        if not self.tracked_in_status:
            return None
        return self.status_games_counter_index()

    @property
    def tracked_games_counter_index(self) -> int | None:
        return self.tracked_status_games_counter_index()

    def tracked_status_completed_counter_index(self) -> int | None:
        if not self.tracked_in_status:
            return None
        return self.status_completed_counter_index()

    @property
    def tracked_completed_counter_index(self) -> int | None:
        return self.tracked_status_completed_counter_index()

    def __str__(self) -> str:
        return self.to_string()


def normalize_quest_level_text(value: str | None) -> str:
    level = QuestLevel.try_parse(value)
    if level is None:
        return ""
    return level.to_string()



from __future__ import annotations

import random
from collections.abc import Callable
from typing import Literal

from grim.assets import PaqTextureCache
from grim.audio import AudioState
from grim.config import CrimsonConfig
from grim.console import ConsoleState
from grim.geom import Vec2
from grim.raylib_api import rl
from grim.view import ViewContext

from ..debug import debug_enabled
from ..game_modes import GameMode
from ..net.debug_log import lan_debug_log
from ..net.rollback_resync_v5 import (
    ModeStateSnapshotV2,
    RushRuntimeSnapshotV2,
    RushStateSnapshotV2,
)
from ..replay import Replay, ReplayHeader, ReplayRecorder, ReplayStatusSnapshot
from ..replay.checkpoints import DEFAULT_CHECKPOINT_SAMPLE_RATE
from ..replay.types import normalize_weapon_usage_counts
from ..sim.bootstrap import BOOTSTRAP_KIND_TERRAIN_V1, run_terrain_bootstrap
from ..sim.sessions import (
    DeterministicSession,
    DeterministicSessionTick,
    RushSpawnState,
    rush_input_transform,
    rush_mid_step,
)
from ..ui.cursor import draw_menu_cursor
from ..ui.hud import HudRenderContext, draw_hud_overlay, hud_flags_for_game_mode
from ..ui.perk_menu import load_perk_menu_assets
from ..weapon_runtime import weapon_assign_player
from ..weapons import WeaponId
from .base_gameplay_mode import (
    BaseGameplayMode,
    LanFramePolicy,
    LanSession,
    LanStepAction,
)
from .components.highscore_record_builder import build_highscore_record_for_game_over

WORLD_SIZE = 1024.0
RUSH_WEAPON_ID = WeaponId.ASSAULT_RIFLE
RUSH_FORCED_AMMO = 30.0

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_ERROR_COLOR = rl.Color(240, 80, 80, 255)

RushSessionFactory = Callable[..., DeterministicSession]


class RushMode(BaseGameplayMode):
    def __init__(
        self,
        ctx: ViewContext,
        *,
        texture_cache: PaqTextureCache | None = None,
        config: CrimsonConfig | None = None,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: random.Random | None = None,
        session_factory: RushSessionFactory = DeterministicSession,
    ) -> None:
        super().__init__(
            ctx,
            world_size=WORLD_SIZE,
            default_game_mode_id=GameMode.RUSH,
            demo_mode_active=False,
            quest_fail_retry_count=0,
            hardcore=False,
            texture_cache=texture_cache,
            config=config,
            console=console,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._ui_assets = None
        self._replay_recorder: ReplayRecorder | None = None
        self._session_factory = session_factory
        self._spawn_state = RushSpawnState()
        self._sim_session: DeterministicSession | None = self._new_sim_session()

    def _enforce_rush_loadout(self) -> None:
        for player in self.sim_world.players:
            if player.weapon.weapon_id != RUSH_WEAPON_ID:
                weapon_assign_player(player, RUSH_WEAPON_ID, state=self.state)
            # Native `rush_mode_update` forces assault rifle + 30 ammo every frame.
            player.weapon.ammo = float(RUSH_FORCED_AMMO)

    def _new_sim_session(self) -> DeterministicSession:
        self._spawn_state = RushSpawnState()
        spawn = self._spawn_state
        return self._session_factory(
            world=self.sim_world.world_state,
            world_size=float(self.world_size),
            damage_scale_by_type=self.sim_world.damage_scale_by_type,
            fx_queue=self.render_resources.fx_queue,
            fx_queue_rotated=self.render_resources.fx_queue_rotated,
            game_mode=GameMode.RUSH,
            perk_progression_enabled=False,
            detail_preset=5,
            gore_disabled=0,
            game_tune_started=bool(self.sim_world.game_tune_started),
            clear_fx_queues_each_tick=False,
            finalize_post_render_lifecycle=True,
            elapsed_uses_raw_dt=True,
            mid_step_hook=lambda ctx: rush_mid_step(ctx, spawn),
            before_step_hook=self._enforce_rush_loadout,
            input_transform=rush_input_transform,
        )

    def open(self) -> None:
        super().open()
        self._ui_assets = load_perk_menu_assets(self._assets_root)
        self._reset_gameplay_frame_clock()
        self._reset_lan_capture_clock()

        status = self.state.status
        base_status = self.save_status
        sim_unlock_index = int(status.quest_unlock_index) if status is not None else 0
        sim_unlock_index_full = int(status.quest_unlock_index_full) if status is not None else 0
        status_unlock_index = int(base_status.quest_unlock_index) if base_status is not None else int(sim_unlock_index)
        status_unlock_index_full = (
            int(base_status.quest_unlock_index_full)
            if base_status is not None
            else int(sim_unlock_index_full)
        )
        quest_unlock_index = int(sim_unlock_index)
        bootstrap = run_terrain_bootstrap(
            self.state.rng,
            quest_unlock_index=int(quest_unlock_index),
            width=int(self.world_size),
            height=int(self.world_size),
            layers=3,
        )
        lan_debug_log(
            "terrain_bootstrap",
            mode="RushMode",
            lan_enabled=bool(self._lan_enabled),
            lan_role=str(self._lan_role),
            status_quest_unlock_index=int(status_unlock_index),
            status_quest_unlock_index_full=int(status_unlock_index_full),
            sim_quest_unlock_index=int(sim_unlock_index),
            sim_quest_unlock_index_full=int(sim_unlock_index_full),
            quest_unlock_index=int(quest_unlock_index),
            seed_before=int(bootstrap.seed_before),
            seed_after=int(bootstrap.seed_after),
            selection_draws=int(bootstrap.selection_draws),
            stamping_draws=int(bootstrap.stamping_draws),
            terrain_base=int(bootstrap.terrain_ids[0]),
            terrain_overlay=int(bootstrap.terrain_ids[1]),
            terrain_detail=int(bootstrap.terrain_ids[2]),
            terrain_seed=int(bootstrap.terrain_seed),
        )
        self.apply_bootstrap_terrain(
            terrain_ids=bootstrap.terrain_ids,
            seed=bootstrap.terrain_seed,
            layers=3,
        )
        self._sim_session = self._new_sim_session()
        self._enforce_rush_loadout()
        weapon_usage_counts = normalize_weapon_usage_counts(
            status.data.get("weapon_usage_counts") if status is not None else None,
        )
        status_snapshot = ReplayStatusSnapshot(
            quest_unlock_index=int(status.quest_unlock_index) if status is not None else 0,
            quest_unlock_index_full=int(status.quest_unlock_index_full)
            if status is not None
            else 0,
            weapon_usage_counts=weapon_usage_counts,
        )
        record_replay = (not bool(self._lan_enabled)) or str(self._lan_role) == "host"
        if record_replay:
            self._replay_recorder = ReplayRecorder(
                ReplayHeader(
                    game_mode_id=GameMode.RUSH,
                    seed=int(self.state.rng.state),
                    bootstrap_kind=BOOTSTRAP_KIND_TERRAIN_V1,
                    bootstrap_seed=int(self._bootstrap_seed),
                    tick_rate=int(self._gameplay_tick_rate()),
                    quest_fail_retry_count=int(self.quest_fail_retry_count),
                    hardcore=bool(self.hardcore),
                    preserve_bugs=bool(self.state.preserve_bugs),
                    detail_preset=int(self._deterministic_detail_preset()),
                    gore_disabled=int(self._deterministic_gore_disabled()),
                    world_size=float(self.world_size),
                    player_count=len(self.sim_world.players),
                    status=status_snapshot,
                ),
            )
            self._replay_checkpoints_sample_rate = int(DEFAULT_CHECKPOINT_SAMPLE_RATE)
        else:
            self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

    def close(self) -> None:
        if self._ui_assets is not None:
            self._ui_assets = None
        self._sim_session = None
        super().close()

    def _handle_input(self) -> None:
        if self._game_over_active:
            if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
                self._action = "back_to_menu"
                self.close_requested = True
            return

        if (not bool(self._lan_enabled)) and rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._paused = not self._paused

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._action = "open_pause_menu"
            return

    def _enter_game_over(self) -> None:
        if self._game_over_active:
            return

        game_mode_id = GameMode(self.config.game_mode)
        record = build_highscore_record_for_game_over(
            state=self.state,
            player=self.player,
            survival_elapsed_ms=int(self._session_elapsed_ms()),
            creature_kill_count=int(self.creatures.kill_count),
            game_mode_id=game_mode_id,
        )

        self._game_over_record = record
        self._game_over_ui.open()
        self._game_over_active = True
        self._save_replay()

    def _replay_checkpoint_elapsed_ms(self) -> float:
        return self._session_elapsed_ms()

    def _replay_claimed_stats_complete(self) -> bool:
        return bool(self._game_over_active)

    def _replay_claimed_stats_elapsed_ms(self) -> int:
        return int(self._session_elapsed_ms())

    def _replay_output_basename(self, *, stamp: str, replay: Replay) -> str:
        _ = replay
        kills = int(self.creatures.kill_count)
        return f"rush_{stamp}_kills{kills}"

    def _lan_mode_name(self) -> Literal["rush"]:
        return "rush"

    def _lan_match_session(self) -> DeterministicSession | None:
        return self._sim_session

    def _lan_frame_policy(self) -> LanFramePolicy:
        return LanFramePolicy(
            prepare_frame=self._rush_prepare_lan_frame,
            on_tick_applied=self._rush_on_tick_applied,
        )

    def _rush_prepare_lan_frame(
        self,
        role: str,
        dt_ui_ms: float,
        session: LanSession,
        dt_tick: float,
    ) -> bool:
        _ = role, dt_ui_ms, dt_tick
        session.detail_preset = int(self._deterministic_detail_preset())
        session.gore_disabled = int(self._deterministic_gore_disabled())
        return True

    def _rush_on_tick_applied(
        self,
        tick: DeterministicSessionTick,
        frame_tick_index: int | None,
        dt_tick: float,
    ) -> LanStepAction:
        _ = tick, dt_tick
        elapsed_ms = self._session_elapsed_ms()
        spawn_cooldown_ms = float(self._spawn_state.spawn_cooldown_ms)
        if frame_tick_index is not None:
            self._store_net_runtime_snapshot(
                snapshot=RushStateSnapshotV2(
                    tick_index=int(frame_tick_index),
                    replay_state=self._net_replay_snapshot_state(),
                    runtime_state=RushRuntimeSnapshotV2(
                        elapsed_ms=elapsed_ms,
                        spawn_cooldown_ms=spawn_cooldown_ms,
                        kill_count=int(self.creatures.kill_count),
                    ),
                ),
            )
        if not self._any_player_alive():
            self._enter_game_over()
            return "stop_after_finalize"
        return "continue"

    def _apply_resync_snapshot(self, snapshot: ModeStateSnapshotV2) -> None:
        if not isinstance(snapshot, RushStateSnapshotV2):
            return
        rs = snapshot.runtime_state
        if self._sim_session is not None:
            self._sim_session.elapsed_ms = float(rs.elapsed_ms)
        self._spawn_state.spawn_cooldown_ms = float(rs.spawn_cooldown_ms)
        self.creatures.kill_count = int(rs.kill_count)

    def update(self, dt: float) -> None:
        frame = self._begin_mode_update(float(dt))
        if frame is None:
            return

        if self._game_over_active:
            self._update_game_over_ui(float(frame.dt))
            return

        if bool(self._lan_enabled) and self._lan_runtime is not None:
            self._update_lan_match(dt=float(frame.dt), dt_ui_ms=0.0)
            return

        any_alive = self._any_player_alive()
        sim_dt = float(frame.dt) if ((not self._paused) and any_alive) else 0.0
        session = self._sim_session

        if self._lan_wait_gate_active():
            self._reset_gameplay_frame_clock()
            return
        if sim_dt <= 0.0:
            self._reset_gameplay_frame_clock()
            if not any_alive:
                self._enter_game_over()
            return
        if session is None:
            return

        tick_dt = float(self._gameplay_tick_dt(session=session))

        def _on_tick(tick, tick_index: int | None) -> bool:
            _ = tick_index
            action = self._rush_on_tick_applied(tick, None, tick_dt)
            return action != "continue"

        def _on_checkpoint(tick_index: int, tick) -> None:
            self._record_replay_checkpoint_from_tick(
                tick_index=int(tick_index),
                tick=tick,
            )

        self._run_deterministic_session_ticks(
            dt_frame=float(sim_dt),
            session=session,
            recorder=self._replay_recorder,
            on_tick=_on_tick,
            on_checkpoint=_on_checkpoint,
        )

    def _draw_game_cursor(self) -> None:
        mouse_pos = self._ui_mouse
        cursor_tex = self._ui_assets.cursor if self._ui_assets is not None else None
        draw_menu_cursor(
            self.render_resources.particles_texture,
            cursor_tex,
            pos=mouse_pos,
            pulse_time=float(self._cursor_pulse_time),
        )

    def draw(self) -> None:
        self._draw_world(
            draw_aim_indicators=(not self._game_over_active),
            entity_alpha=self._world_entity_alpha(),
        )
        self._draw_screen_fade()

        hud_bottom = 0.0
        if (not self._game_over_active) and self._hud_assets is not None:
            hud_flags = hud_flags_for_game_mode(self._config_game_mode_id())
            self._draw_target_health_bar()
            hud_bottom = draw_hud_overlay(
                HudRenderContext(
                    assets=self._hud_assets,
                    state=self._hud_state,
                    font=self._small,
                    show_health=hud_flags.show_health,
                    show_weapon=hud_flags.show_weapon,
                    show_xp=hud_flags.show_xp,
                    show_time=hud_flags.show_time,
                    show_quest_hud=hud_flags.show_quest_hud,
                    small_indicators=self._hud_small_indicators(),
                ),
                player=self.player,
                players=self.sim_world.players,
                bonus_hud=self.state.bonus_hud,
                elapsed_ms=self._session_elapsed_ms(),
                frame_dt_ms=self._last_dt_ms,
            )

        if debug_enabled() and (not self._game_over_active):
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            line = float(self._ui_line_height())
            self._draw_ui_text(
                f"rush: t={self._session_elapsed_ms() / 1000.0:6.1f}s",
                Vec2(x, y),
                UI_TEXT_COLOR,
            )
            self._draw_ui_text(f"kills={self.creatures.kill_count}", Vec2(x, y + line), UI_HINT_COLOR)
            y_extra = y + line * 2.0
            if self._paused:
                self._draw_ui_text("paused (TAB)", Vec2(x, y_extra), UI_HINT_COLOR)
                y_extra += line
            if self.player.health <= 0.0:
                self._draw_ui_text("game over", Vec2(x, y_extra), UI_ERROR_COLOR)
                y_extra += line
            self._draw_lan_debug_info(x=x, y=y_extra, line_h=line)

        if self._game_over_active:
            self._draw_game_cursor()
            if self._game_over_record is not None:
                self._game_over_ui.draw(
                    record=self._game_over_record,
                    banner_kind=self._game_over_banner,
                    hud_assets=self._hud_assets,
                    mouse=self._ui_mouse_pos(),
                )
        self._draw_lan_wait_overlay()
